__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import textwrap
import pandas as pd
from umiche.deduplicate.method.trimer.Collapse import Collapse
from umiche.fastq.Reader import Reader as rfastq
from umiche.trim.Reader import Reader as trimreader
from umiche.util.Hamming import Hamming

from umiche.simu.Parameter import Parameter as params

from umiche.util.Reader import Reader as freader
from umiche.util.Writer import Writer as fwriter
from umiche.util.Console import Console


class selfHealing:
 
    def __init__(
            self,
            scenario,
            param_fpn=None,
            verbose=False,

            **kwargs,
    ):
        self.scenario = scenario
        self.kwargs = kwargs

        self.params = params(param_fpn=param_fpn)

        if 'fastq_fp' not in self.kwargs.keys():
            self.fastq_fp = self.params.fastq_fp
        else:
            self.fastq_fp = self.kwargs['fastq_fp']

        if 'umi_ref_fpn' not in self.kwargs.keys():
            self.umi_ref_fpn = self.params.umi_ref_fpn
        else:
            self.umi_ref_fpn = self.kwargs['umi_ref_fpn']

        self.rfastq = rfastq()
        self.collapse = Collapse()
        self.trimreader = trimreader()
        self.freader = freader()
        self.fwriter = fwriter()

        self.console = Console()
        self.console.verbose = verbose

        self.df_umi_ref_lib = self.freader.generic(df_fpn=self.umi_ref_fpn)
        self.df_umi_ref_lib = pd.DataFrame.from_dict({
            i: e for i, e in enumerate(self.df_umi_ref_lib[0].values)
        }, orient='index')
        self.console.print(self.df_umi_ref_lib)
        self.umi_monomer_ref_dict = self.df_umi_ref_lib[0].apply(lambda x: ''.join([i[0] for i in textwrap.wrap(x, 3)])).to_dict()
        self.console.print(self.umi_monomer_ref_dict)

    def rea(self, ) -> str:
        df_stat = pd.DataFrame()

        for perm_num_i in range(self.params.fixed['permutation_num']):
            for id, scenario_i in enumerate(self.params.varied[self.scenario]):

                read_stime = time.time()
                if self.scenario == 'pcr_nums':
                    self.fn_mark = str(id)  # scenario_i
                elif self.scenario == 'pcr_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'ampl_rates':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_deps':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_lens':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_nums':
                    self.fn_mark = str(id)
                else:
                    self.fn_mark = str(scenario_i)
                self.fn_prefix = self.params.file_names[self.scenario] + self.fn_mark

                names, seqs, _, _ = self.rfastq.fromgz(
                    fastq_fpn=self.fastq_fp + self.fn_prefix + '.fastq.gz',
                )
                self.console.print('=========>read time: {:.3f}s'.format(time.time() - read_stime))

                df_fastq = self.trimreader.todf(names=names, seqs=seqs)

                mono_corr_stime = time.time()
                df_fastq['umi_mono_corr'] = df_fastq['umi'].apply(lambda x: self.collapse.majority_vote(x))
                # df_fastq['umi_mono_corr'] = df_fastq['umi'].apply(lambda x: self.correct(x))
                self.console.print('=========>mono_corr time: {:.3f}s'.format(time.time() - mono_corr_stime))

                hm_stime = time.time()
                df_stat['umi_hm' + str(id)] = df_fastq.apply(lambda x: Hamming().general(
                    x['umi_mono_corr'],
                    self.umi_monomer_ref_dict[x['umi#']],
                ), axis=1)
                self.console.print('=========>Hamming time: {:.3f}s'.format(time.time() - hm_stime))
                print(df_stat['umi_hm' + str(id)])

        self.fwriter.generic(
            df=df_stat,
            sv_fpn=to('data/simu/umi/seq_errs/trimer/trimmed/dasd1111.txt'),
        )
        return 'Finished.'

    def stat(self, x):
        return Hamming().general(x['umi_mono_corr'], self.umi_monomer_ref_dict[x['umi#']])


if __name__ == "__main__":
    from umiche.path import to

    p = selfHealing(
        # scenario='pcr_nums',
        # scenario='pcr_errs',
        scenario='seq_errs',
        # scenario='ampl_rates',
        # scenario='umi_lens',
        # scenario='seq_deps',
        # scenario='umi_nums',

        # umi_ref_fpn=to('data/simu/umi/trimer/seq_errs/umi.txt'),
        # fastq_fp=to('data/simu/umi/trimer/seq_errs/trimmed/'),

        param_fpn=to('data/trimer.yml'),

        verbose=True, # False True
    )

    print(p.rea())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

from umiche.deduplicate.io.Stat import Stat


def read_bam(
        bam_fpn : str,
        verbose : bool =True
):
    from umiche.bam.Reader import Reader as alireader
    return alireader(
        bam_fpn=bam_fpn,
        verbose=verbose,
    )


def stat(
        scenarios,
        methods,
        param_fpn,
        verbose : bool =True
):
    return Stat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=param_fpn,
        verbose=verbose,
    )


if __name__ == "__main__":
    from umiche.path import to

    # bam = read_bam(
    #     bam_fpn="/mnt/d/Document/Programming/Python/umiche/umiche/data/simu/umi/trimer/seq_errs/permute_0/trimmed/seq_err_17.bam",
    #     verbose=True,
    # )
    # print(bam.todf(tags=['PO']))

    stat_p = stat(
        scenarios={
            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'ampl_rates': 'Amplification rate',
            'umi_lens': 'UMI length',
            'seq_deps': 'Sequencing depth',
        },
        methods={
            # 'unique': 'Unique',
            # 'cluster': 'Cluster',
            # 'adjacency': 'Adjacency',
            'directional': 'Directional',
            # 'dbscan_seq_onehot': 'DBSCAN',
            # 'birch_seq_onehot': 'Birch',
            # 'aprop_seq_onehot': 'Affinity Propagation',
            'mcl': 'MCL',
            # 'mcl_val': 'MCL-val',
            # 'mcl_ed': 'MCL-ed',
        },
        param_fpn=to('data/params.yml'),
        verbose=True,
    )

    print(stat_p.df_dedup)
    print(stat_p.df_trace_cnt)
    print(stat_p.df_inflat_exp)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.util.Reader import Reader as freader


class DedupSingle:

    def __init__(
            self,
            df,
            df_melt,
    ):
        # sns.set(font="Verdana")
        sns.set(font="Helvetica")
        sns.set_style("ticks")

        self.freader = freader()
        self.df = df
        self.df_melt = df_melt
        print(self.df)
        print(self.df_melt)

    def jointplot(
            self,
            method,
    ):
        ppp = sns.jointplot(
            x=self.df[self.df['method'] == method]['mean'].values,
            y=self.df[self.df['method'] == 'Directional']['mean'].values,
            kind="reg",
            color="crimson",
            label='asd',
        )
        ppp.ax_joint.plot([0, 8.5], [0, 8.5], 'grey', linewidth=2, alpha=1)
        ppp.set_axis_labels(
            r'$\frac{N_e-N_t}{N_t}$' + ' (' + method + ')',
            r'$\frac{N_e-N_t}{N_t}$' + ' (Directional)',
            fontsize=14,
        )
        ppp.ax_joint.text(8, 8.5, "confidence interval", horizontalalignment='right', size='medium', color='crimson',)
        ppp.ax_joint.text(3, 5.5, "regression", horizontalalignment='left', size='medium', color='crimson',)
        ppp.ax_joint.text(5.5, 5, "baseline", horizontalalignment='left', size='medium', color='black',)
        sns.despine(right=True, top=True)
        plt.tight_layout()
        plt.show()

    def jointgrid(
            self,
            method='MCL-ed',
    ):
        # fig, ax = plt.subplots()
        self.df_melt = self.df_melt.rename(columns={'value': r'$\frac{N_e-N_t}{N_t}$'})

        self.df_melt['Sequencing error'] = self.df_melt['Sequencing error'].astype(float)
        self.df_melt = self.df_melt.loc[self.df_melt['Sequencing error'] <5.0e-03]
        print(self.df_melt)
        g = sns.JointGrid(
            data=self.df_melt[self.df_melt['method'] == method],
            x="Sequencing error",
            y=r'$\frac{N_e-N_t}{N_t}$',
            marginal_ticks=True,
        )

        # Create an inset legend for the histogram colorbar
        cax = g.figure.add_axes([.20, .55, .02, .2])
        g.ax_joint.set_xlabel('Sequencing error', fontsize=16)
        g.ax_joint.set_ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=18)
        # Add the joint and marginal histogram plots
        g.plot_joint(
            sns.kdeplot,
            discrete=(True, False),
            cmap="light:#03012d", pmax=.8, cbar=True, cbar_ax=cax,
        )

        g.ax_joint.set_title(method, fontsize=14)
        xticks = self.df_melt['Sequencing error'].unique()
        print(xticks[xticks < 5.0e-03])
        g.ax_joint.set_xticks(np.linspace(xticks[xticks <5.0e-03][0], xticks[xticks <5.0e-03][-1], 4))
        # g.ax_joint.set_xticklabels(xticks[xticks <5.0e-03])
        # g.ax_joint.set_xticklabels([1e-05, '', '', '', '', '', '', '', 0.001, 0.0025, 0.005, 0.0075, 0.01, 0.025])
        g.ax_joint.set_xticklabels(np.linspace(xticks[xticks <5.0e-03][0], xticks[xticks <5.0e-03][-1], 4))
        plt.setp(g.ax_joint.get_xticklabels(), rotation=45)

        g.plot_marginals(
            sns.histplot,
            element="step",
            color="#03012d",
        )
        # sns.despine(left=True, bottom=True)
        plt.tight_layout()
        plt.show()

    def strip(self, ):
        fig, ax = plt.subplots()
        sns.despine(bottom=True, left=True)
        cc = [
            'tab:green',
            'tab:blue',
            'crimson',
        ]
        sns.stripplot(
            x="value",
            y="Sequencing error",
            hue='method',
            palette=cc,
            data=self.df_melt,
            dodge=True,
            alpha=.25,
            zorder=1,
        )
        sns.pointplot(
            x="value",
            y="Sequencing error",
            hue='method',
            data=self.df_melt,
            dodge=.8 - .8 / 3,
            join=False,
            palette=cc,
            markers="d",
            scale=.75,
            ci=None,
        )
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(
            handles[len(handles)//2:],
            labels[len(handles)//2:],
            title='',
            handletextpad=0,
            columnspacing=1,
            loc="upper right",
            ncol=3,
            frameon=True,
            fontsize=12,
        )
        ax.set_xlabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)
        ax.set_ylabel('Sequencing error rate', fontsize=16)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.15,
            left=0.15,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.show()

    def stackedbar(self, ):
        fig, ax = plt.subplots(figsize=(4, 5))

        self.df_direc = self.df[self.df['method'] == 'Directional']
        self.df_mcl_val = self.df[self.df['method'] == 'MCL-val']
        self.df_mcl_ed = self.df[self.df['method'] == 'MCL-ed']

        self.df_mcl_val["dmean"] = self.df_direc["mean"] - self.df_mcl_val["mean"]
        self.df_mcl_ed["dmean"] = self.df_direc["mean"] - self.df_mcl_ed["mean"]

        # self.df_mcl_val["dmean"] = np.exp(self.df_direc["mean"] - self.df_mcl_val["mean"])
        # self.df_mcl_ed["dmean"] = np.exp(self.df_direc["mean"] - self.df_mcl_ed["mean"])

        sns.set_color_codes("pastel")
        sns.barplot(
            x="dmean",
            y="metric",
            data=self.df_mcl_val,
            label="dFC_ed",
            color="b",
        )
        sns.set_color_codes("muted")
        sns.barplot(
            x="dmean",
            y="metric",
            data=self.df_mcl_ed,
            label="dFC_val",
            color="b",
        )

        # plt.fill_between(self.df_mcl_val["dmean"], self.df_mcl_val["metric"])
        ax.legend(ncol=2, loc="upper right", frameon=True, fontsize=12)
        # ax.set(ylabel="", xlabel="Automobile collisions per billion miles")
        ax.set_ylabel('Sequencing error', fontsize=16)
        ax.set_xlabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)
        sns.despine(left=True, bottom=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.14,
            left=0.24,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.show()

    def errorbar(self, ):
        # plt.rc('text', usetex=True)
        # plt.rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
        fig, ax = plt.subplots()
        # ax.plot(
        #     file_n301.index,
        #     n301_means,
        #     label='Full TrainData dataset',
        #     alpha=0.9,
        #     linewidth=3.0,
        #     # s=,
        #     c='royalblue'
        # )
        cc = [
            'tab:green',
            'tab:blue',
            'crimson',
        ]

        df_gp = self.df.groupby(by=['method'])
        gp_keys = df_gp.groups.keys()

        for i, met in enumerate(gp_keys):
            df_met = df_gp.get_group(met)

            ax.errorbar(
                x=df_met.index,
                y=df_met['mean'],
                yerr=[df_met['mean-min'], df_met['max-mean']],
                fmt='o',
                alpha=0.6,
                ecolor=cc[i],
                color=cc[i],
                linewidth=1,
                elinewidth=1,
                capsize=2,
                markersize=6,
                label=met,
            )
            # print(df_met['metric'])
            ax.set_xticks(np.arange(df_met['metric'].shape[0]))
            ax.set_xticklabels(df_met['metric'], fontsize=8, rotation=45)
        # ax.set_xlabel('Sequencing error rate', fontsize=12)
        # ax.set_xlabel('Amplification rate', fontsize=12)
        # ax.set_xlabel('UMI length', fontsize=12)
        ax.set_xlabel('Polymerase error rate', fontsize=12)
        ax.set_ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)
        # ax.set_title(DEFINE['title'], fontsize=12)
        sns.despine(right=True, top=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.16,
            left=0.11,
            # left=0.12,
            right=0.95,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.legend(fontsize=11, loc='upper left')
        plt.show()

    def errorband(self, ):
        fig, ax = plt.subplots()
        cc = [
            'tab:green',
            'tab:blue',
            # 'dimgray',
            'crimson',
        ]
        df_gp = self.df.groupby(by=['method'])
        gp_keys = df_gp.groups.keys()

        for i, met in enumerate(gp_keys):
            df_met = df_gp.get_group(met)
            ax.errorbar(
                x=df_met.index,
                y=df_met['mean'],
                yerr=[df_met['mean-min'], df_met['max-mean']],
                fmt='o',
                alpha=0.7,
                ecolor=cc[i],
                color=cc[i],
                linestyle='-',
                linewidth=2,
                elinewidth=0.5,
                capsize=2,
                markersize=3,
                label=met,
            )
            ax.plot(df_met.index, df_met['mean'] - df_met['mean-min'], color=cc[i], linewidth=0.1, alpha=0.1)
            ax.plot(df_met.index, df_met['max-mean'] + df_met['mean'], color=cc[i], linewidth=0.1, alpha=0.1)
            ax.fill_between(
                df_met.index,
                df_met['mean'] - df_met['mean-min'],
                df_met['max-mean'] + df_met['mean'],
                alpha=0.1,
                color=cc[i],
            )
            ax.set_xticks(df_met['metric'].index)
            ax.set_xticklabels(df_met['metric'], fontsize=8, rotation=45)

        # ax.set_xlabel('Sequencing error rate', fontsize=12)
        ax.set_xlabel('Amplification rate', fontsize=12)
        ax.set_ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=14)
        # ax.set_title(DEFINE['title'], fontsize=12)
        sns.despine(right=True, top=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.16,
            left=0.13,
            right=0.95,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.legend(fontsize=11)
        plt.show()


if __name__ == "__main__":
    from umiche.path import to
    from umiche.deduplicate.io.Stat import Stat as dedupstat

    scenarios = {
        # 'pcr_nums': 'PCR cycle',
        # 'pcr_errs': 'PCR error',
        'seq_errs': 'Sequencing error',
        # 'ampl_rates': 'Amplification rate',
        # 'umi_lens': 'UMI length',
        # 'seq_deps': 'Sequencing depth',
    }
    methods = {
        # 'unique': 'Unique',
        # 'cluster': 'Cluster',
        # 'adjacency': 'Adjacency',
        'directional': 'Directional',
        # 'dbscan_seq_onehot': 'DBSCAN',
        # 'birch_seq_onehot': 'Birch',
        # 'aprop_seq_onehot': 'Affinity Propagation',
        'mcl': 'MCL',
        'mcl_val': 'MCL-val',
        'mcl_ed': 'MCL-ed',
    }
    dedupstat = dedupstat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=to('data/params.yml'),
    )
    df_dedup = dedupstat.df_dedup
    df_dedup_perm_melt = dedupstat.df_dedup_perm_melt

    p = DedupSingle(
        df=df_dedup,
        df_melt=df_dedup_perm_melt,
    )
    # print(p.strip())
    # print(p.jointplot(
    #     # method='MCL',
    #     method='MCL-val',
    #     # method='MCL-ed',
    # ))
    print(p.jointgrid(
        # method='Directional',
        # method='MCL',
        # method='MCL-val',
        method='MCL-ed',
    ))
    print(p.stackedbar())
    # print(p.errorbar())
    # print(p.errorband())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import h5py
import json
import scanpy


class reader:

    def __init__(self, ):
        pass

    def anndata(self, h5_fpn):
        return scanpy.read_h5ad(h5_fpn)

    def gcmatLit(self, h5_fpn):
        from scipy.sparse import csr_matrix, coo_matrix, csc_matrix
        with h5py.File(h5_fpn, 'r') as f:
            dict_genes = json.loads(f['default/dict_genes'][()])
            dict_cells = json.loads(f['default/dict_cells'][()])
            dict_cell_cls = json.loads(f['default/dict_cell_cls'][()])
            dict_genes = {int(k): v for k, v in dict_genes.items()}
            dict_cells = {int(k): v for k, v in dict_cells.items()}
            # print(len(dict_genes))
            # print(f['default/gc_mat_vals'][()])
            # print(f['default/gc_mat_ids'][()])
            # print(f['default/gc_mat_id_pointers'][()])
            gc_mat = csr_matrix(
                (f['default/gc_mat_vals'][()], f['default/gc_mat_ids'][()], f['default/gc_mat_id_pointers'][()]),
                shape=(len(dict_cells), len(dict_genes))
            ).toarray()
            # print(gc_mat)
            # print(dict_genes)
            # print(dict_cells)
            # print(dict_cell_cls)
        return {
            'gc_mat': gc_mat,
            'dict_genes': dict_genes,
            'dict_cells': dict_cells,
            'dict_cell_cls': dict_cell_cls,
        }


if __name__ == "__main__":
    p = reader()
    print(p.anndata(h5_fpn='../../../../notebook/sc/Lechner/file.h5ad'))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.stats as st
from umiche.simu.Parameter import Parameter as params
from umiche.deduplicate.io.Stat import Stat as dedupstat
from umiche.plot.gadget.Transmitter import Transmitter as transmitter

from umiche.util.Reader import Reader as freader


class DedupMultipleTrimerSetCover:

    def __init__(
            self,
            scenarios: Dict,
            methods: Dict,
            by: str = 'pn0',
            is_trans: bool = False,
            umi_gt_cnt: int = 50,
            param_fpn: str = None,
    ):
        self.scenarios = scenarios
        self.methods = methods
        self.by = by
        self.is_trans = is_trans
        self.umi_gt_cnt = umi_gt_cnt
        self.param_fpn = param_fpn
        self.freader = freader()
        self.params = params(param_fpn=self.param_fpn)

        self.dedupstat = dedupstat(
            scenarios=self.scenarios,
            methods=self.methods,
            param_fpn=self.param_fpn,
            is_trans=self.is_trans,
        )

        self.df_dedup = self.dedupstat.df_dedup_set_cover_len

        # print(self.df_dedup)
        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def violinplot(
            self,
            num_img_row=1,
            num_img_col=1,
            start=0,
            end=-1,
    ):
        y_dict = {}
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]
                y_dict[method_formal] = []
                self.df_sce_met[self.by].apply(lambda x: y_dict[method_formal].append([int(i) for i in x.split(';')]))
        x_labels = self.df_sce_met['metric'].values[start:end]

        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(17, 5.5), sharey=False, sharex=False)
        # palette = sns.color_palette("Paired")[2:]
        palette = sns.color_palette("Set3")[3:]
        BLACK = "#282724"
        GREY_DARK = "#747473"
        displacement = [0.4, -0.4]

        for i, (method_formal, y_2d) in enumerate(y_dict.items()):
            y_2d = y_2d[start:end]

            position_init = [t for t in np.arange(len(y_2d) * 2)][::2]
            positions = [t - displacement[i] for t in position_init]
            print(positions)

            x_2d = [[t] * len(d) for t, d in zip(positions, y_2d)]
            x_jittered = [x_2d + st.t(df=6, scale=0.04).rvs(len(x_2d)) for x_2d in x_2d]

            violins = ax.violinplot(
                y_2d,
                positions=positions,
                widths=0.8,
                bw_method="silverman",
                showmeans=False,
                showmedians=False,
                showextrema=False
            )

            for pc in violins["bodies"]:
                pc.set_facecolor("none")
                pc.set_edgecolor(BLACK)
                pc.set_linewidth(1.4)
                pc.set_alpha(1)

            medianprops = dict(
                linewidth=4,
                color=GREY_DARK,
                solid_capstyle="butt"
            )
            boxprops = dict(
                linewidth=2,
                color=GREY_DARK
            )

            ax.boxplot(
                y_2d,
                widths=0.4,
                positions=positions,
                showfliers=False,
                showcaps=False,
                medianprops=medianprops,
                whiskerprops=boxprops,
                boxprops=boxprops,
                # showmeans=True,
            )

            for u, (x, y) in enumerate(zip(x_jittered, y_2d)):
                ax.scatter(
                    x,
                    y,
                    s=20,
                    color=palette[i],
                    label=method_formal if u == 0 else None,
                    # facecolor='none',
                    alpha=0.4,
                )

        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_linewidth(2)
        ax.spines['left'].set_linewidth(2)
        ax.set_xticks(position_init)
        ax.set_xticklabels(x_labels, fontsize=12)
        ax.set_xlabel('Sequencing error rate', fontsize=18)
        ax.set_ylabel('Count', fontsize=18)
        plt.rcParams["legend.markerscale"] = 2
        ax.legend(frameon=False, fontsize=14)

        plt.subplots_adjust(
            top=0.96,
            bottom=0.12,
            left=0.05,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        # plt.savefig('./solve_line.eps', format='eps')
        plt.show()
        return

    def violinplot_sgl(
            self,
            num_img_row=1,
            num_img_col=1,
            start=0,
            end=-1,
    ):
        y_dict = {}
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]
                y_dict[method_formal] = []
                self.df_sce_met[self.by].apply(lambda x: y_dict[method_formal].append([int(i) for i in x.split(';')]))
        x_labels = self.df_sce_met['metric'].values[start:end]

        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(6, 5), sharey=False, sharex=False)
        # palette = sns.color_palette("Paired")[2:]
        palette = sns.color_palette("Set3")[3:]
        BLACK = "#282724"
        GREY_DARK = "#747473"
        displacement = [0.4, -0.4]

        for i, (method_formal, y_2d) in enumerate(y_dict.items()):
            y_2d = y_2d[start:end]

            position_init = [t for t in np.arange(len(y_2d) * 2)][::2]
            positions = [t - displacement[i] for t in position_init]
            print(positions)

            x_2d = [[t] * len(d) for t, d in zip(positions, y_2d)]
            x_jittered = [x_2d + st.t(df=6, scale=0.04).rvs(len(x_2d)) for x_2d in x_2d]

            violins = ax.violinplot(
                y_2d,
                positions=positions,
                widths=0.4,
                bw_method="silverman",
                showmeans=False,
                showmedians=False,
                showextrema=False
            )

            for pc in violins["bodies"]:
                pc.set_facecolor("none")
                pc.set_edgecolor(BLACK)
                pc.set_linewidth(1.4)
                pc.set_alpha(1)

            medianprops = dict(
                linewidth=4,
                color=GREY_DARK,
                solid_capstyle="butt"
            )
            boxprops = dict(
                linewidth=2,
                color=GREY_DARK
            )

            ax.boxplot(
                y_2d,
                widths=0.2,
                positions=positions,
                showfliers=False,
                showcaps=False,
                medianprops=medianprops,
                whiskerprops=boxprops,
                boxprops=boxprops,
                # showmeans=True,
            )

            for u, (x, y) in enumerate(zip(x_jittered, y_2d)):
                ax.scatter(
                    x,
                    y,
                    s=20,
                    color=palette[i],
                    label=method_formal if u == 0 else None,
                    # facecolor='none',
                    alpha=0.4,
                )

        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_linewidth(2)
        ax.spines['left'].set_linewidth(2)
        ax.set_xticks(position_init)
        ax.set_xticklabels(x_labels, fontsize=12)
        ax.set_xlabel('Sequencing error rate', fontsize=18)
        ax.set_ylabel('Count', fontsize=18)
        plt.rcParams["legend.markerscale"] = 2
        ax.legend(frameon=False, fontsize=14, loc='lower center', bbox_to_anchor=(0.5, 1.0))

        plt.subplots_adjust(
            top=0.82,
            bottom=0.12,
            left=0.12,
            right=0.98,

            # top=0.82,
            # bottom=0.135,
            # left=0.15,
            # right=0.98
        )
        # plt.savefig('./solve_line.eps', format='eps')
        plt.show()
        return

    def line(
            self,
            num_img_row=1,
            num_img_col=1,
    ):
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(6.5, 5.5), sharey=False, sharex=False)
        palette = sns.color_palette("Paired")[:]
        # palette = sns.color_palette("Set3")[3:]
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False

                # ax[int(n / num_img_col), n % num_img_col].errorbar(
                ax.errorbar(
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    yerr=[self.df_sce_met['mean-min'], self.df_sce_met['max-mean']],
                    fmt='o',
                    # alpha=0.6,
                    # ecolor=palette[j],
                    color=palette[j],
                    linestyle='-',
                    linewidth=2,
                    elinewidth=1,
                    capsize=3,
                    markersize=6,
                    markeredgewidth=1.2,
                    markeredgecolor=palette[j],
                    markerfacecolor='none',
                    label=method_formal,
                )

                self.line_gadget(
                    ax=ax,
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    line_color=palette[j],
                    # label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    # title='{}'.format(scenario_formal),
                    title_fs=16,
                    x_label='{}'.format(scenario_formal), # 'Position'
                    y_label='Count',
                    x_label_rotation=30,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(self.df_sce_met.index.shape[0])).tolist(),
                    x_ticklabels=self.df_sce_met['metric'].values.tolist(),
                    x_ticklabel_fs=12,
                    x_label_fs=18,
                    y_label_fs=18,
                    legend_fs=14,
                    legend_loc=None,
                )
                ax.spines['bottom'].set_linewidth(2)
                ax.spines['left'].set_linewidth(2)

        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.16,
            left=0.12, # 0.14 0.12
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        # plt.savefig('./umitools mets.eps', format='eps')
        # plt.savefig('./mv+sc.eps', format='eps')
        plt.show()
        return

    @transmitter(type="line", task=None)
    def line_gadget(*args, **kwargs):
        return kwargs

    @transmitter(type="line_scatter", task=None)
    def line_scatter_gadget(*args, **kwargs):
        return kwargs

if __name__ == "__main__":
    from umiche.path import to

    p = DedupMultipleTrimerSetCover(
        scenarios={
            # 'pcr_nums': 'PCR cycle',
            # 'pcr_errs': 'PCR error rate',
            'seq_errs': 'Sequencing error rate',
            # 'ampl_rates': 'Amplification rate',
            # 'umi_lens': 'UMI length',
            # 'seq_deps': 'Sequencing depth',
        },

        methods={
            'set_cover_mono_len_split_by_mv': 'monomer UMIs generated by spMV',
            'set_cover_mono_len_split_to_all': 'monomer UMIs generated by spALL',
            # 'set_cover_multi_len_split_by_mv': 'homotrimer UMIs removed by spMV',
            # 'set_cover_multi_len_split_to_all': 'homotrimer UMIs removed by spALL',
        },

        is_trans=True, # False True

        param_fpn=to('data/params_trimer.yml'),
    )

    # p.violinplot(
    #     start=0,
    #     end=-1,
    #
    #     # start=10,
    #     # end=-1,
    #
    #     # start=0,
    #     # end=10,
    # )

    # p.violinplot_sgl(
    #     start=15,
    #     end=-1,
    # )

    p.line()


import pandas as pd
import numpy as np
import seaborn as sns
import pdb
from matplotlib import pyplot as plt
from matplotlib.patches import Patch
import pdb

def build_comparison_table(dfs, names, fillna=None, column_name='Scenarios'):
    """ Simple function to combine dataframes in a multi-index comparison table

    Parameters
    ----------
    dfs : list of pandas dataframes to be compared
    names : list of names to assign to each dataframe
        Labels to distinguish each dataframe's respective data in the compiled comparison table
    fillna: None (default), or float
        Value with which to replace any potential NaN instance
    column_name : string
        Name of multi-index level that holds the `names` values.

    Returns
    -------
    comp: dataframe
        A comparison dataframe

    """

    for i, df in enumerate(dfs):
        # Add the name of each dataframe for disambiguation
        df.insert(0, column_name, names[i], allow_duplicates=True)

    # Concatenate
    comp = pd.concat(dfs, sort=False)
    comp.sort_index(inplace=True)
    comp.set_index(column_name, append=True, inplace=True)

    if fillna is not None:
        comp.fillna(fillna, inplace=True)

    return comp

def plot_grouped_stackedbars(df, ix_categories, ix_entities_compared, norm='max', err_pos=None, err_neg=None,
                             palette_def=('pastel', 'deep', 'dark'), width=0.3, figsize=(8, 4), xaxis_label=''):
    """ Grouped stacked-bars for both comparison and contribution analysis

    This plot groups bars, representing the total scores of different entities
    [ix_entities_compared] in terms of multiple comparison categories
    [ix_categories], while at the same time breaking down these total scores
    into contribution analysis.

    For example, compare two vehicles (ix_entities_compared) in terms of
    climate change and resource depletion impacts (ix_categories), and break
    down their total impacts in terms of multiple lifecycle stages, such as the
    contribution of vehicle production, use phase and end-of-life treatment.

    The function uses a color polette to distinguish between lifecycle stages,
    and increasingly darker variants of this palette (from pastel, to deep, to
    dark, by default) to distinguish between the technologies/scenarios
    being compared (ix_entities_compared).

    All colour palettes must be defined explicitly. For comparisons involving
    many compared entities (>3), it may be more convenient to use
    `plot_grouped_stackedbar_wlargegroups()`, which automatically generates the
    shading gradient. Unfortunately, this function does not allow for the
    definition of error margins (confidence intervals).


    Parameters
    ----------

    df : pandas multi-index dataframe, as generated from build_comparsison_table()

        The DataFrame must be multi-index, with one index level indicating the
        comparison category/criteria, and another level indicating the entities
        being compared. All columns must represent an element of the
        contribution analysis

        Important: ALL columns in the dataframe must be relevant for the
        contribution analysis, except those that are singled out as defining
        confidence intervals (`err_pos`, `err_neg`). All other columns should
        be removed or used as indexes

    ix_categories : string
        The name of the index level that holds the categories/criteria for the
        comparison. For example, types of life cycle impacts

    ix_entities_compared : string
        The name of the index level that holds the entities being compared,
        such as competing products, technologies, or scenarios

    norm : None or string {'max' | index of reference entity }
        If Norm is None, the stacked bars are not being normalized

        If norm == 'max' (default): Within each comparison categories, the
        different entities are normalized relative to the entity with the
        largest score.

        If norm is the index of a specific reference entity, all other
        entities are normalized relative to that one.

    err_neg, err_pos: None, or string
        The name of the column that holds the negative and positive errors
        associated with the total sum of each row of df values
        If None, no error bars will be drawn

    palette_def: tupple of matplotlib or seaborn "categorical" palette definitions
        These palettes should present the same colors, but with different
        lightness levels, forming a gradient from lightest to darkest.

    width : float
        The width of the bars.

    figsize : tuple
        The width and height of figure

    xaxis_label : string
        The label/title of the x-axis

    Returns
    -------
    ax, fig : matplotlib axes and figure

    See Also
    --------
    Depends on internal functions _normalize_impacts() and _generate_legend()

    """

    # Alternative approach (maybe simpler for just two scenarios):
    # https://stackoverflow.com/questions/40121562/clustered-stacked-bar-in-python-pandas

    # Hardcoded
    edgecolor = 'k'
    transparent = (0, 0, 0, 0)

    # Test that there are not too many entities compared.
    if len(df.index.get_level_values(ix_entities_compared).unique()) > 3:
        print("Warning. Too many entities being compared. Please use `plot_grouped_stackedbars_wlargegroups()` instead.")
        return None

    # Normalize
    if norm is not None:
        if norm == 'max':
            df = _normalize_impacts(df, ix_categories, ix_entities_compared,
                                    donotsumbutnormalize=(err_neg, err_pos))
        else:
            df = _normalize_impacts(df, ix_categories, ix_entities_compared, ref=norm,
                                    donotsumbutnormalize=(err_neg, err_pos))

    # Initializations. Creates two subplots, the first one (ax) for the actual figure, and the second (ax2) as dummy to
    # display the legend without this legend overlapping with the plot but still within the bounds of the figure.
    # width_ratios is dummy ratio to make second plot very small
    fig, (ax, ax2) = plt.subplots(1, 2, gridspec_kw = {'width_ratios':[100, 1]},
                                  facecolor='white', figsize=figsize)
    # make dummy axis disapear
    ax2.get_yaxis().set_ticks([])
    ax2.axis('off')

    # All compared entities, in order of appearance
    all_entities = df.index.get_level_values(ix_entities_compared).unique()
    all_contributions = [i for i in df.columns if i not in [err_neg, err_pos]]

    # Determine number of entities, portions, and colors
    n_entities_compared = len(all_entities)
    n_palettes = len(palette_def)

    # Define palettes
    palettes = []
    if n_entities_compared <= n_palettes:
        for i in range(n_palettes - n_entities_compared, n_palettes):
            palettes += [sns.color_palette(palette_def[i])]
    else:
        palettes = None

    # Loop over all entities compared
    for i, ent in enumerate(all_entities):

        # Subset of contribution data for this loop
        sub = df.xs(ent, axis=0, level=ix_entities_compared)[all_contributions]

        if err_pos:
            # Error bar data fro this loop, if applicable
            err = df.xs(ent, axis=0, level=ix_entities_compared)[[err_neg, err_pos]].values.T
        else:
            err = None

        # Plot horizontal bar
        sub.plot.barh(ax=ax, stacked=True, position=i, width=width, zorder=-1,
                      color=sns.color_palette(palettes[i]), edgecolor=edgecolor, label='_nolegend_')

        # Plot over this bar with a transparent bar, to add the confidence interval
        sub.sum(1).plot.barh(ax=ax, position=i, width=width, color=transparent, xerr=err, label='_nolegend_')

        if i == n_entities_compared - 1:

            # Generate the legend complement explaining about shading
            legend_elements = _generate_legend(all_entities)

            # Integrate in legend and plot legend
            handles, labels = ax.get_legend_handles_labels()
            handles2 = handles[-len(all_contributions):] + legend_elements

    # Legend definition (hidden in first subplot, but displayed in second)
    ax.get_legend().remove()
    ax2.legend(handles=handles2, loc='center left', bbox_to_anchor=(0, 0.5))

    # Axis title
    ax.xaxis.set_label_text(xaxis_label)
    ax.yaxis.set_label_text('')

    # Rescale
    ax.autoscale()  # Important to not have the bars come right up to th edge of figure
    fig.tight_layout() # Important to not have labels and legend extend beyond figure

    return ax, fig


def plot_grouped_stackedbars_wlargegroups(df, ix_categories, ix_entities_compared, norm='max', orient='h',
                                         palette_def=('pastel', 'deep', 'dark')):
    """ Grouped stacked-bars for both comparison and contribution analysis

    Group bars, representing the total scores of different compared entities
    [ix_entities_compared] around multiple comparison categories
    [ix_categories], while at the same time breaking down these totals into
    contribution analysis.

    For example, compare two vehicles (ix_entities_compared) in terms of
    climate change and resource depletion (ix_categories), and break down their
    total impacts in terms of multiple lifecycle stages, such as vehicle
    production, use phase and end-of-life treatment.

    Parameters
    ----------

    df : DataFrame, following a specific format
        The DataFrame must be multi-index, with one index level indicating the
        comparison category/criteria, and another level indicating the entities
        being compared. All columns must represent an element of the
        contribution analysis

    ix_categories : string
        The name of the index level that holds the categories/criteria for the
        comparison. For example, types of life cycle impacts

    ix_entities_compared : string
        The name of the index level that holds the entities being compared,
        such as competing products, technologies, or scenarios

    norm : None or string {'max' | index of reference entity }
        If Norm is None, the stacked bars are not being normalized

        If norm == 'max', within each comparison categories, the different
        entities are normalized relative to the largest contributor

        If norm is the index of a reference entity, all other
        entities are normalized relative to that one.

    orient : string {'h' | 'v'}
        Whether to have horizontal ('h') or vertical ('v') bar graph

    palette_def: tupple of matplotlib or seaborn "categorical" palette definitions
        These palettes should present the same colors, but with different
        lightness levels, forming a gradient from lightest to darkest.


    Returns
    -------

    fig : matplotlib figure object
    ax :  matplotlib axis object


    See Also
    --------

    This function is really the sequential application of three internal
    functions: `_normalize_impacts()`, `_calc_cumsum_tidy_df`, and
    `_plot_grouped_stackedbars_from_tidycumsum`.

    """

    # Whether to normalize dataframe rows, and if so relative to either the
    # rowsum of a specific entity (from index) or the entity with the maximum
    # rowsum for each category
    if norm is not None:
        if norm == 'max':
            df = _normalize_impacts(df,
                                    ix_categories=ix_categories,
                                    ix_ref_level=ix_entities_compared)
        else:
            df = _normalize_impacts(df,
                                    ix_categories=ix_categories,
                                    ix_ref_level=ix_entities_compared,
                                    ref=norm)

    # Calculate the row-wise cummulative sum in the dataframe, as a trick to
    # make a stacked bar graph
    df = _calc_cumsum_tidy_df(df, var_name='stages')

    # Make grouped stacked bar contribution-comparison graph
    ax, fig = _plot_grouped_stackedbars_from_tidycumsum(df,
                                                        categories=ix_categories,
                                                        stacked_portions='stages',
                                                        values='value',
                                                        entities_compared=ix_entities_compared,
                                                        orient=orient,
                                                        palette_def=palette_def)
    return ax, fig


def plot_stochastic_comparison(scen1, scen2, name1, name2, palette='deep'):
    """ Plot the probability that scenario 1 or 2 have lower impacts

    Parameters
    ----------
    scen1, scen2 : Pandas DataFrames
        With impacts as indexes, defining the comparison categories, and with
        stochastic iterations as columns

    name1, name2 : string
        Names to identify the scenarios to be compared, i.e. those in scen1 and
        scen2

    palette : seaborn or matplotlib color palette to use


    Returns
    -------
    ax, fig : matplotlib axes and figure

    """

    # Initialize and get dimensions
    n_iter = scen1.shape[1]
    prob_better = pd.DataFrame(index=scen1.index, columns=[name1, name2])

    # Perform calculation as numpy array (50x faster)
    scen2 = scen2.reindex_like(scen1)
    scen1 = scen1.values
    scen2 = scen2.values

    # calculate probability of scen1 being better than scen2 for all impact categories
    prob_better.loc[:, name1] = -1 * ((scen1 < scen2).sum(1) + 0.5 * (scen1 == scen2).sum(1)) / n_iter * 100

    # And vice versa
    # note: we calculate the probability of scenario2 < scenario1 independently.
    #       this could be done more efficiently with `100% - prob_better.loc[:, name1]`, 
    #       but it has the potential of hiding failures. For example if a whole row is zero by mistake 
    #       (NaN issue, etc.), we want both scenarios at 0%, not accidently ascribe 100% to scenario2
    #       as the residual
    prob_better.loc[:, name2] = ((scen1 > scen2).sum(1) + 0.5 * (scen1 == scen2).sum(1)) / n_iter * 100

    # Plotting
    fig = plt.figure(facecolor='white')
    ax = plt.subplot(111)
    prob_better.plot.barh(ax=ax,
                          stacked=True,
                          color=sns.color_palette(palette),
                          edgecolor='k')

    # Make the x-axis stretch from 100% Scenario 1 to 100% Scenario 2
    ax.set_xlim(-100, 100)

    # Vertical line at 0%
    ax.axvline(color='k', linewidth=1)

    # Get rid of negative values in axis labels, turn into positive integers
    xticklabels = [str(abs(int(i))) for i in ax.get_xticks()]
    ax.set_xticklabels(xticklabels)

    # Label and legend
    ax.set_ylabel('')
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2),  ncol=2, fancybox=True)

    fig.tight_layout() # Important to not have labels and legend extend beyond figure
    return ax, fig


def _normalize_impacts(df, ix_categories, ix_ref_level, ref=None, donotsumbutnormalize=('err_neg', 'err_pos')):
    """ Express the score of each entity as percentage of one specific entity

    Parameters
    ----------

    df: Multi-index Pandas DataFrame
        Indexes on categories (e.g. different environmental impacts) and on
        compared entities (e.g., technologies, scenarios). Columns are
        contributions to the total score of each category (e.g., lifecycle
        stages).

    ix_categories: string
        Name of index level containing the categories in terms of which the
        compared entities are quantified

    ix_ref_level: string
        Name of index levels containing the compared entities

    ref : None or string
        If None, normalize relative to the entity with the highest score in
        each category

        If string, should be the index of the entity against which to normalize
        all others

    donotsumbutnormalize: tuple
        All columns that do not contribute to the total but should also be
        normalized

    Returns
    -------

    df: Normalized dataframe

    """

    # Identify all non-error columns

    stages = list(set(df.columns) - set(donotsumbutnormalize))

    if ref:
        # Normalize values for all entities in all categories, relative to one
        # selected reference entity (ref)
        ref_imp = df.xs(ref, axis=0, level=ix_ref_level)[stages].sum(axis=1)

    else:
        # Normalize values for all entities based on the entity with the
        # highest row-sum for each category
        grouped_levels = list({i for i in df.index.names} - {ix_categories})
        ref_imp = df[stages].sum(axis=1).reset_index(grouped_levels, drop=True).groupby(ix_categories).max()

    # Reindex the resulting reference values, to allow for "broadcast" division
    ref_imp = ref_imp.reindex(df.index, level=ix_categories)

    # Divide the dataframe row-wise, for a normalized result
    return df.divide(ref_imp, axis=0) * 100


def _calc_cumsum_tidy_df(df, var_name='stages'):
    """ Prepares a dataframe for grouped stacked-bar plotting

    Turn rows into a cummulative sum, to cheat our way into plotting stacked
    bars

    To facilitate plotting in seaborn, `melt()` the dataframe such that it
    becomes a tidy dataframe.

    Parameters
    ----------

    df : Multi-index Pandas dataframe
        As generated by by `build_comparison_table()`

    var_name: string, optional
        Name to give to the new column in the tidy dataframe that holds all
        melted original columns

        For example, if the columns of the original `df` defined life-cycle
        stages, the newly formed column might be named 'stages'

    Return
    ------

    Tidy dataframe with values expressed as cummulative sums

    """

    # Turn each row in to a sequential cummulative sum
    df = df.cumsum(axis=1)

    # Save names of indexes, and reset
    ix_names = df.index.names
    df = df.reset_index()

    # Turn into a tidy format
    return df.melt(id_vars=ix_names, var_name=var_name)


def _plot_grouped_stackedbars_from_tidycumsum(cumsum_df, categories, stacked_portions, values, entities_compared,
                                              orient='h', palette_def=('pastel', 'deep', 'dark')):
    """ Plotting function behind `plot_grouped_stackedbar_comparison()`

      Mode 2) If the number of entities is greater than the number of defined palettes,
      a single palette is used to define the 'color' arguments of seaborn
      barplots, leaving seaborn to determine the lightness gradient. The first
      entity in the comparison is typically too light, but a large number of
      bars together makes it clear to interpret regardless.



    Parameters
    ----------

    cumsum_df : Pandas DataFrame, as generated from _calc_cumsum_tidy_df()

    categories : string
        The name of the column that holds the categories/criteria for the
        comparison. For example, types of life cycle impacts

    entities_compared : string
        The name of the column that holds the entities being compared,
        such as competing products, technologies, or scenarios

    stacked_portions : string
        the name of the column that holds the sections of the stacked bar

    values : float
        The name of the column with the values that determine the size of each
        section in stacked bars.

    orient : string {'h' | 'v'}
        Whether to have horizontal ('h') or vertical ('v') bar graph

    palette_def: tupple of matplotlib or seaborn "categorical" palette definitions
        These palettes should present the same colors, but with different
        lightness levels, forming a gradient from lightest to darkest.

    """

    # Alternative approach (maybe simpler for just two scenarios):
    # https://stackoverflow.com/questions/40121562/clustered-stacked-bar-in-python-pandas

    # Initializations
    fig = plt.figure(facecolor='white')
    ax = plt.subplot(111)
    legend_elements = []

    # Determine number of entities, portions, and colors
    n_entities_compared = len(set(cumsum_df[entities_compared]))
    n_stacked_portions = len(set(cumsum_df[stacked_portions]))
    n_palettes = len(palette_def)

    # Define colors, in case this function is run as "Mode 2"
    colors = sns.color_palette(palette_def[-1], n_colors=n_stacked_portions)

    # Define palettes if this function is run as "Mode 1"
    palettes = []
    if n_entities_compared <= n_palettes:
        for i in range(n_palettes - n_entities_compared, n_palettes):
            palettes += [sns.color_palette(palette_def[i], n_colors=n_stacked_portions)]
    else:
        palettes = None

    # Swap x and y axes, for vertical or horizontal bars, depending
    if orient == 'h':
        x = values
        y = categories
    else:
        x = categories
        y = values

    # Loop over all stacked proportions, to plot each in order, on over the
    # other, creating the illusion of stacked bars
    for i, j in enumerate(cumsum_df[stacked_portions].unique()):

        # Select the right data for this loop
        bo = cumsum_df[stacked_portions] == j
        g = cumsum_df.loc[bo]

        # Sort out the right color or palette, depending on number of entities
        # compared relative ot the number of predefined palettes
        if n_entities_compared > n_palettes:
            color = colors[i]
            palette = None
        else:
            color = None
            palette = [p[i] for p in palettes]

        ax = sns.barplot(data=g, x=x, y=y, hue=entities_compared, color=color, palette=palette,
                         zorder=n_stacked_portions - i, edgecolor="k", ci=None)

        if i == 0:
            legend_elements = _generate_legend(g[entities_compared].unique())
        try:
            legend_elements += [Patch(facecolor=palette[-1], label=j), ]
        except TypeError:
            legend_elements += [Patch(facecolor=color, label=j), ]

    # Remove context-dependent stuff
    ax.legend_.remove()  # remove the redundant legends
    ax.xaxis.set_label_text('')
    ax.yaxis.set_label_text('')

    legend_elements.reverse()
    plt.legend(handles=legend_elements,
               bbox_to_anchor=(1, 0.5),
               bbox_transform=fig.transFigure,
               loc="center left")

    return ax, fig


def _generate_legend(entities):
    """ Generate legend explaining shading in grouped_stackbar graphs, based on order and number of entities compared

    Handles either two entities (light & dark), three entities (light medium and dark), or more.
    In this last case, it generates a single entry stating the order of the gradient (lightest to darkest).

    Parameters
    ----------
    entities : list
        List of strings of names of the entities being compared, in the order that they are processed

    Returns
    -------
    legend_element : matplotlib legend
    """

    legend_elements = []

    if len(entities) > 3:
        legend_elements += [Patch(edgecolor='black',
                                  facecolor='white',
                                  label="Gradient: {} (lightest), {}, {} (darkest)".format(entities[0],
                                                                                           ', '.join(entities[1: -1]),
                                                                                           entities[-1]))]

    else:

        legend_elements += [Patch(edgecolor='black',
                                  facecolor='white',
                                  label='light colors: ' + entities[0])]
        if len(entities) == 3:
            legend_elements += [Patch(edgecolor='black', 
                                      facecolor='white', 
                                      label='medium colors: ' + entities[1])]

        legend_elements += [Patch(edgecolor='black', 
                                  facecolor='white', 
                                  label='dark colors: ' + entities[-1])]
    return legend_elements


# It might be pertinent to define and lighten/darken colormaps ourselves:
# https://scipy-cookbook.readthedocs.io/items/Matplotlib_ColormapTransformations.html



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

import pandas as pd
import collections
from umiche.deduplicate.method.trimer.Collapse import Collapse
from umiche.util.Console import Console


class SetCover:

    def __init__(
            self,
            verbose=False,
    ):
        import pysam

        self.collapse = Collapse()

        self.console = Console()
        self.console.verbose = verbose

    def umi_greedy_li(
            self,
            umi_dict,
    ):
        """

        Parameters
        ----------
        umi_arr

        Returns
        -------

        """
        # print(umi_arr)
        merged_umis = {}
        merged_umis_idx = {}
        trimer_umi_list = [*umi_dict.keys()]
        # @@ trimer_umi_list[0]
        # ['GGGTTTGTGACCCCCTGTAAATTTCCCCGGAAAGTG',
        # 'GGGAAATTTTTTGTTCTCAAAGGGCAAGGGAAATTT',
        # ...,
        # 'AAAGGGAAACCCAAATTTGGGTTTTCGTTTCCTTTT',]
        mono_umi_set_list = [*umi_dict.values()]
        # @@ mono_umi_set_list
        # [{'GTGCCTATCGAG', 'GTGACGATCGAT', ..., 'GTGCCGATCGAT'},
        # {'GATTGCAGAGAT', 'GATTTTAGCGAT', ..., 'GATTGCAGCGAT'},
        # ...,
        # {'AGACATGTGTTT', 'AGACATGTCTTT', ..., 'AGACATGTTTCT'}]
        n_merged = len(mono_umi_set_list)
        mono_umi_set_list_remaining = mono_umi_set_list
        n_steps = 0
        # print(left_umis)
        while n_merged > 1:
            mono_umi_to_trimer_id_dict = dict()
            for ii, mono_umi_set in enumerate(mono_umi_set_list_remaining):
                # print(len(left_umis))
                for mono_umi in mono_umi_set:
                    if mono_umi in mono_umi_to_trimer_id_dict:
                        mono_umi_to_trimer_id_dict[mono_umi].append(ii)
                    else:
                        mono_umi_to_trimer_id_dict[mono_umi] = [ii]
            # @@ mono_umi_to_trimer_id_dict
            # {'AGATGGACGGCA': [4523, 4989, 5607],
            # 'AGTAGGATGGCA': [4523, 5725],
            # ...,
            # 'GAAGCGAATTTC': [6521]}
            umi_counts = {kk: len(vv) for kk, vv in mono_umi_to_trimer_id_dict.items()}

            df_umi = pd.DataFrame({
                'mono_umi': umi_counts.keys(),
                'val_cnt': umi_counts.values(),
            }).sort_values(
                by='val_cnt',
                ascending=False,
            ).reset_index(drop=True)
            # print(df_umi)

            n_merged = df_umi.val_cnt[0]

            if n_merged > 1:
                merged_umis_idx[n_steps] = mono_umi_to_trimer_id_dict[df_umi.mono_umi[0]]
                merged_umis[n_steps] = df_umi.mono_umi[0]
                # print(merged_umis_idx)
                # fix umis in path
                mono_umi_set_list_remaining_tmp = []
                for rr_idx, mono_left_umi_set in enumerate(mono_umi_set_list_remaining):
                    # print(rr_idx)
                    # print(mono_left_umi_set)
                    if rr_idx not in merged_umis_idx[n_steps]:
                        # print(rr_idx, merged_umis_idx[n_steps])
                        # break
                        mono_umi_set_list_remaining_tmp.append(mono_left_umi_set)
                # print(mono_umi_set_list_remaining_tmp)
                # print(len(mono_umi_set_list_remaining_tmp))
                # break
                mono_umi_set_list_remaining = mono_umi_set_list_remaining_tmp
                if len(mono_umi_set_list_remaining) == 0:
                    break
                n_steps += 1
            else:
                break
        # print(df_umi.loc[df_umi['val_cnt'] == 1])
        # print(df_umi.loc[df_umi['val_cnt'] == 1].shape)
        # print(mono_umi_set_list_remaining)
        # solutiom_unique_umis = set(merged_umis.values())
        # solution_umis = [ii for ii, cc in enumerate(mono_umi_set_list) if len(cc.intersection(solutiom_unique_umis)) > 0]
        shortest_path = len(mono_umi_set_list) - sum([len(ii) - 1 for ii in merged_umis_idx.values()])
        # print('shortest_path: ', shortest_path)
        return (shortest_path)

    def count_li(
            self,
            inbam,
            tag,
            sep="_",
    ):
        tab = dict()
        # count = 0
        n_tag = 0
        with pysam.AlignmentFile(inbam) as bf:
            for i, r in enumerate(bf):
                if r.has_tag(tag) is True:

                    key = ('tag', r.get_tag(tag))

                    tab.setdefault(key, []).append(r.qname.split(sep)[1])
                    print(tab)
                    n_tag += 1
                else:
                    pass

            self.console.print("======>The total number of input reads is ".format(i + 1))
            self.console.print("======>The total number of input reads with XT tag is ".format(n_tag))

            trimer_with_mono = collections.defaultdict(list)

            n = 0
            for tag, trimer_list in tab.items():
                # print(str(tag))
                # print(len(trimer_list))
                if len(trimer_list) == 1:
                    count = 1
                    n += count

                    monocmi = self.collapse.cmi_li(trimer_list).pop()
                    # print(monocmi)
                    keys = str(tag) + '_' + str(trimer_list)
                    trimer_with_mono[keys].append(monocmi)

                else:
                    # corrected_cmis = set(tuple(self.collapse.ShuangLiCollapseCMI(uu)) for uu in trimer_list)
                    corrected_cmis = set(tuple(self.collapse.split_by_mv(uu)) for uu in trimer_list)
                    # print(corrected_cmis)

                    if len(corrected_cmis) == 1:
                        count = 1
                        n += count

                        for trimer in trimer_list:
                            keys = '_'.join(tag) + '_' + str(trimer)
                            # print(keys)
                            trimer_with_mono[keys].append(corrected_cmis)

                    else:
                        trimer_to_combos = {tri: self.collapse.split_to_all(tri) for tri in trimer_list}
                        # print(trimer_to_combos)

                        # count = self.umi_greedy([self.collapse.ShuangLiCollapseUMI(uu) for uu in ['TTTCCCTTTAAAGGGTTTGGGCCCCCC', 'TTGCCGTTTAAAGGGTTTGGGCCCCCC', 'TAGCAGTTGATAGGGTTTGGGCCCCCC']])
                        # count = self.umi_greedy([self.collapse.ShuangLiCollapseUMI(uu) for uu in trimer_list])
                        count = self.umi_greedy(trimer_to_combos)
                        # count = self.umi_greedy([self.collapse.split_to_all(uu) for uu in trimer_list])

                        # trimer_with_mono = {str('_'.join(tag)) + "_" + str(tri): mono for tri, mono in new_trimer_to_combos.items()}

                        n += count
        return n

    def greedy1(
            self,
            multimer_list,
    ):
        """

        Parameters
        ----------
        umi_arr

        Returns
        -------

        """
        umi_dict = {multimer_umi: self.collapse.split_to_all(multimer_umi) for multimer_umi in multimer_list}

        merged_umis = {}
        merged_mono_umi_dict = {}
        merged_umis_idx = {}
        trimer_umi_to_id_map = {trimer_umi: k for k, trimer_umi in enumerate(umi_dict.keys())}
        trimer_id_to_umi_map = {k: trimer_umi for k, trimer_umi in enumerate(umi_dict.keys())}
        # @@ [*umi_dict.keys()]
        # ['GGGTTTGTGACCCCCTGTAAATTTCCCCGGAAAGTG',
        # 'GGGAAATTTTTTGTTCTCAAAGGGCAAGGGAAATTT',
        # ...,
        # 'AAAGGGAAACCCAAATTTGGGTTTTCGTTTCCTTTT',]
        mono_umi_set_list = [*umi_dict.values()]
        # @@ mono_umi_set_list
        # [{'GTGCCTATCGAG', 'GTGACGATCGAT', ..., 'GTGCCGATCGAT'},
        # {'GATTGCAGAGAT', 'GATTTTAGCGAT', ..., 'GATTGCAGCGAT'},
        # ...,
        # {'AGACATGTGTTT', 'AGACATGTCTTT', ..., 'AGACATGTTTCT'}]
        n_merged = len(mono_umi_set_list)
        mono_umi_set_list_remaining = umi_dict
        n_steps = 0
        # print(left_umis)
        while n_merged > 1:
            mono_umi_to_trimer_id_dict = dict()
            for multimer_umi, mono_umi_set in mono_umi_set_list_remaining.items():
                # print(len(left_umis))
                for mono_umi in mono_umi_set:
                    if mono_umi in mono_umi_to_trimer_id_dict:
                        mono_umi_to_trimer_id_dict[mono_umi].append(trimer_umi_to_id_map[multimer_umi])
                    else:
                        mono_umi_to_trimer_id_dict[mono_umi] = [trimer_umi_to_id_map[multimer_umi]]
            # print(mono_umi_to_trimer_id_dict)
            # @@ mono_umi_to_trimer_id_dict
            # {'AGATGGACGGCA': [4523, 4989, 5607],
            # 'AGTAGGATGGCA': [4523, 5725],
            # ...,
            # 'GAAGCGAATTTC': [6521]}
            umi_counts = {kk: len(vv) for kk, vv in mono_umi_to_trimer_id_dict.items()}

            df_umi = pd.DataFrame({
                'mono_umi': umi_counts.keys(),
                'val_cnt': umi_counts.values(),
            }).sort_values(
                by='val_cnt',
                ascending=False,
            ).reset_index(drop=True)
            # print(df_umi)

            n_merged = df_umi.val_cnt[0]

            if n_merged > 1:
                merged_umis_idx[n_steps] = mono_umi_to_trimer_id_dict[df_umi.mono_umi[0]]
                merged_umis[n_steps] = df_umi.mono_umi[0]

                merged_mono_umi_dict[df_umi.mono_umi[0]] = trimer_id_to_umi_map[mono_umi_to_trimer_id_dict[df_umi.mono_umi[0]][0]]
                print(merged_mono_umi_dict)

                # fix umis in path
                mono_umi_set_list_remaining_tmp = {}
                # print(len(mono_umi_set_list_remaining))
                for multimer_umi, mono_left_umi_set in mono_umi_set_list_remaining.items():
                    # print(multimer_umi)
                    # print(mono_left_umi_set)
                    if trimer_umi_to_id_map[multimer_umi] not in merged_umis_idx[n_steps]:
                        # print(multimer_umi, merged_umis_idx[n_steps])
                        # break
                        mono_umi_set_list_remaining_tmp[multimer_umi] = mono_left_umi_set
                # print(mono_umi_set_list_remaining_tmp)
                # print(len(mono_umi_set_list_remaining_tmp))
                # break
                mono_umi_set_list_remaining = mono_umi_set_list_remaining_tmp
                print(len(mono_umi_set_list_remaining))

                if len(mono_umi_set_list_remaining) == 0:
                    break
                n_steps += 1
            else:
                break
        multimer_umi_solved_by_sc = [*merged_mono_umi_dict.values()]
        multimer_umi_not_solved = [*mono_umi_set_list_remaining.keys()]
        print('=========>number of shortlisted multimer UMIs solved by set cover'.format(len(multimer_umi_solved_by_sc)))
        print('=========>number of shortlisted multimer UMIs cannot be solved by set cover'.format(len(multimer_umi_not_solved)))
        # solutiom_unique_umis = set(merged_umis.values())
        # solution_umis = [ii for ii, cc in enumerate(mono_umi_set_list) if len(cc.intersection(solutiom_unique_umis)) > 0]
        shortlisted_multimer_umi_list = multimer_umi_solved_by_sc + multimer_umi_not_solved
        dedup_cnt = len(mono_umi_set_list) - sum([len(ii) - 1 for ii in merged_umis_idx.values()])
        print('dedup_cnt: ', dedup_cnt)
        return dedup_cnt, shortlisted_multimer_umi_list

    def greedy(
            self,
            multimer_list,
            recur_len,
            split_method='split_to_all',
    ):
        """

        Parameters
        ----------
        umi_arr

        Returns
        -------

        """
        if split_method == 'split_to_all':
            split_func = self.collapse.split_to_all
        else:
            split_func = self.collapse.split_by_mv
        umi_dict = {multimer_umi: split_func(
            umi=multimer_umi,
            recur_len=recur_len,
        ) for multimer_umi in multimer_list}
        # print(umi_dict)
        monomer_umi_lens = []
        multimer_umi_lens = []
        merged_mono_umi_dict = {}
        trimer_umi_to_id_map = {trimer_umi: k for k, trimer_umi in enumerate(umi_dict.keys())}
        trimer_id_to_umi_map = {k: trimer_umi for k, trimer_umi in enumerate(umi_dict.keys())}
        # print(trimer_umi_to_id_map)
        # print(trimer_id_to_umi_map)
        # @@ [*umi_dict.keys()]
        # ['GGGTTTGTGACCCCCTGTAAATTTCCCCGGAAAGTG',
        # 'GGGAAATTTTTTGTTCTCAAAGGGCAAGGGAAATTT',
        # ...,
        # 'AAAGGGAAACCCAAATTTGGGTTTTCGTTTCCTTTT',]
        mono_umi_set_list = [*umi_dict.values()]
        # print(mono_umi_set_list)
        # @@ mono_umi_set_list
        # [{'GTGCCTATCGAG', 'GTGACGATCGAT', ..., 'GTGCCGATCGAT'},
        # {'GATTGCAGAGAT', 'GATTTTAGCGAT', ..., 'GATTGCAGCGAT'},
        # ...,
        # {'AGACATGTGTTT', 'AGACATGTCTTT', ..., 'AGACATGTTTCT'}]
        mono_umi_set_list_remaining = umi_dict
        num_steps = 0
        is_empty_set_overlap = False
        while not is_empty_set_overlap:
            # It addresses how many trimer UMIs monomer UMIs can account for
            mono_umi_to_trimer_id_dict = {}
            for multimer_umi, mono_umi_set in mono_umi_set_list_remaining.items():
                for mono_umi in mono_umi_set:
                    if mono_umi in mono_umi_to_trimer_id_dict:
                        mono_umi_to_trimer_id_dict[mono_umi].append(trimer_umi_to_id_map[multimer_umi])
                    else:
                        mono_umi_to_trimer_id_dict[mono_umi] = [trimer_umi_to_id_map[multimer_umi]]
            # @@ mono_umi_to_trimer_id_dict
            # {'GGATTCGGGACT': [5022, 6458], ..., 'TAAAAAGATTAT': [6890], 'TAAAATGACTAT': [6890]}
            monomer_umi_lens.append(len(mono_umi_to_trimer_id_dict))
            monomer_umi_to_cnt_map = {k: len(v) for k, v in mono_umi_to_trimer_id_dict.items()}
            # @@ monomer_umi_to_cnt_map
            # {'GGATTCGGGACT': 2, ..., 'TAAAAAGACTAT': 1, 'TAAAATGATTAT': 1}
            if monomer_umi_to_cnt_map:
                monomer_umi_max = max(monomer_umi_to_cnt_map, key=monomer_umi_to_cnt_map.get)
            else:
                break
            # print(monomer_umi_max)
            # TTAGATGATTAT
            # ...
            # TTTTAAGCTGTC
            # TCCTCTAGTGCC
            if monomer_umi_to_cnt_map[monomer_umi_max] > 1:
                multimer_umi_ids = mono_umi_to_trimer_id_dict[monomer_umi_max]
                multimer_umi_lens.append(len(multimer_umi_ids) - 1)

                # important!!
                # @@ this is where we keep one trimer UMI
                merged_mono_umi_dict[monomer_umi_max] = trimer_id_to_umi_map[mono_umi_to_trimer_id_dict[monomer_umi_max][0]]

                for multimer_umi_id in multimer_umi_ids:
                    mono_umi_set_list_remaining.pop(trimer_id_to_umi_map[multimer_umi_id], None)
                num_steps += 1
                is_empty_set_overlap = False
            else:
                is_empty_set_overlap = True

        multimer_umi_solved_by_sc = [*merged_mono_umi_dict.values()]
        multimer_umi_not_solved = [*mono_umi_set_list_remaining.keys()]
        shortlisted_multimer_umi_list = multimer_umi_solved_by_sc + multimer_umi_not_solved
        self.console.print('=========># of shortlisted multimer UMIs solved by set cover: {}'.format(len(multimer_umi_solved_by_sc)))
        self.console.print('=========># of shortlisted multimer UMIs not solved by set cover: {}'.format(len(multimer_umi_not_solved)))
        self.console.print('=========># of shortlisted multimer UMIs: {}'.format(len(shortlisted_multimer_umi_list)))
        # print(num_steps,212323)
        dedup_cnt = len(mono_umi_set_list) - sum(multimer_umi_lens)
        self.console.print('=========>dedup cnt: {}'.format(dedup_cnt))
        return dedup_cnt, multimer_umi_solved_by_sc, multimer_umi_not_solved, shortlisted_multimer_umi_list, monomer_umi_lens, multimer_umi_lens


if __name__ == "__main__":
    from umiche.path import to

    p = SetCover(
        verbose=True,
    )

    # from umiche.bam.Build import Build as umibuild
    # umibuild = umibuild

    from umiche.bam.Reader import Reader as alireader
    alireader = alireader(bam_fpn="/mnt/d/Document/Programming/Python/umiche/umiche/data/simu/umi/trimer/seq_errs/permute_0/trimmed/seq_err_17.bam", verbose=True)
    df_bam = alireader.todf(tags=['PO'])
    print(df_bam.columns)
    print(df_bam.query_name.apply(lambda x: x.split('_')[1]).values)

    (dedup_cnt,
     multimer_umi_solved_by_sc,
     multimer_umi_not_solved,
     shortlisted_multimer_umi_list,
     monomer_umi_lens,
     multimer_umi_lens) = p.greedy(
        multimer_list=df_bam.query_name.apply(lambda x: x.split('_')[1]).values,
        recur_len=3,
        split_method='split_to_all',
    )
    print(dedup_cnt)
    # print(p.count_li(
    #     inbam=to('data/simu/umi/trimer/seq_errs/permute_0/trimmed/seq_err_17.bam'),
    #     tag='PO',
    # ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.util.Console import Console


class seqRuleOut:

    def __init__(self, read_summary, verbose=True):
        self.read_summary = read_summary
        self.verbose = verbose
        self.console = Console()
        self.console.verbose = self.verbose

    def sequential(self, compo_struct, seq_pos_in_struct):
        """

        Notes
        -----------
        Starting positions of all genomic sequences.

        Example
        -------
        rule_out_struct_dict returns all structures before each key of rule_out_struct_dict: {
            'seq_1': [struct_1, struct_2, ..., struct_n],
            'seq_2': [struct_1, struct_2, ..., struct_n],
            ...,
            'seq_m': [struct_1, struct_2, ..., struct_n],
            }
            e.g., {'seq_1': ['primer_1', 'umi_1'], 'seq_2': []}
            if the structure is 'primer_1+umi_1+seq_1+seq_2+umi_2+primer_2'
            * each key does not count all structures of its preceding keys.

        rule_out_rel_len_dict returns accumulated lengths of all structures in the list w.r.t each
        key of rule_out_struct_dict.

        rule_out_accumu_len_dict returns the starting positions of all UMIs.

        Parameters
        ----------
        compo_struct
            1d list of strings of seq_struct split by +.
        seq_pos_in_struct
            1d list of indices of compo_struct.

        Returns
        -------
        1d dict: {umi_1: int, umi_2: int, ..., umi_n: int}

        """
        self.console.print('======>finding the starting positions of all genomic sequence...')
        rule_out_struct_dict = {}
        for i in range(len(seq_pos_in_struct)):
            if i == 0:
                rule_out_struct_dict['seq_' + str(i + 1)] = compo_struct[:seq_pos_in_struct[i]]
            else:
                rule_out_struct_dict['seq_' + str(i + 1)] = compo_struct[seq_pos_in_struct[i - 1] + 1: seq_pos_in_struct[i]]
        rule_out_rel_len_dict = {}
        for key, val in rule_out_struct_dict.items():
            rule_out_rel_len_dict[key] = 0
            if val == []:
                rule_out_rel_len_dict[key] = 0
            else:
                for j in val:
                    rule_out_rel_len_dict[key] += self.read_summary[j]['len']
        rule_out_accumu_len_dict = {}
        accumu = []
        for key, val in rule_out_rel_len_dict.items():
            accumu.append(val)
            rule_out_accumu_len_dict[key] = sum(accumu)
            accumu.append(self.read_summary[key]['len'])
        for k, v in rule_out_accumu_len_dict.items():
            self.console.print('=========>{} starting position: {}'.format(k, v))
        return rule_out_accumu_len_dict


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

import pandas as pd
from umiche.simu.Parameter import Parameter as params
from umiche.fastq.Reader import Reader as fastqreader
from umiche.util.Console import Console


class Anchor:

    def __init__(
            self,
            scenario,
            param_fpn,
            verbose=False,
    ):
        self.fastqreader = fastqreader()
        self.scenario = scenario
        self.param_fpn = param_fpn
        self.params = params(param_fpn=param_fpn)
        print(self.params.anchor)
        self.verbose = verbose
        self.console = Console()
        self.console.verbose = self.verbose

    def pct_read_captured(self, ):
        without_anchor = {}
        with_anchor = {}
        for permut_i in range(self.params.fixed['permutation_num']):
            self.console.print("===>permutation test {}".format(permut_i))
            without_anchor[permut_i] = []
            with_anchor[permut_i] = []
            for i, criterion in enumerate(self.params.varied['criteria']):
                arr_fastq = self.fastqreader.fromgz(fastq_fpn=self.params.work_dir + '/permute_' + str(permut_i) + "/" + self.scenario + "_" + str(i) + '.fastq.gz')
                df = pd.DataFrame(arr_fastq).T
                self.console.print('======>number of reads at criterion {}: {}'.format(criterion, df.shape[0]))
                df['len'] = df[1].apply(lambda x: len(x))
                without_anchor[permut_i].append(df.loc[df['len'] == 92].shape[0]/df.shape[0])
                df['is_anchor'] = df[1].apply(lambda x: x.__contains__(self.params.anchor['seq1']))
                with_anchor[permut_i].append(df[df['is_anchor']].shape[0]/df.shape[0])
        return with_anchor, without_anchor


if __name__ == "__main__":
    from umiche.path import to

    anchor = Anchor(
        scenario='pcr_del',
        param_fpn=to('data/params_anchor.yml'),
    )



import numpy as np


class mean(object):

    def __init__(self, gc_mat):
        self.gc_mat = gc_mat

    def pergene(self, ):
        return self.gc_mat.mean(axis=0)

    def percell(self, ):
        return self.gc_mat.mean(axis=1)

    def percellcls(self, cls_vec={}):
        if not isinstance(cls_vec, dict):
            raise ValueError
        res = {}
        for k, v in cls_vec.items():
            res[k] = self.gc_mat[v, :].mean(axis=1).tolist()
        return res

    def allcellcls(self, cls_vec={}):
        if not isinstance(cls_vec, dict):
            raise ValueError
        res = {}
        for k, v in cls_vec.items():
            res[k] = self.gc_mat[v, :].mean().tolist()
        return res

    def pergenecls(self, cls):
        pass

    def allgenecls(self, cls):
        pass

    def all(self, ):
        return self.gc_mat.mean()


if __name__ == "__main__":
    gc_mat = np.array([
        [3., 3., 3., 6., 6.],
        [1., 1., 1., 2., 2.],
        [1., 22., 1., 2., 2.]
    ])
    p = mean(gc_mat)
    print(p.pergene())
    print(p.percell())
    print(p.all())

    cls_vec = {0: [0, 2], 1: [1]}
    print(p.percellcls(cls_vec))
    print(p.allcellcls(cls_vec))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"



from umiche.trim.Template import Template


def template(
        params,
):
    trimmer = Template(params)
    df = trimmer.todf()
    trimmer.togz(df)
    return df


if __name__ == "__main__":
    from umiche.path import to

    params = {
        'umi_1': {
            'len': 10,
        },
        'umi_2': {
            'len': 4,
        },
        'bc_1': {
            'len': 2,
        },
        'read_struct': 'umi_1',
        # 'read_struct': 'umi_1+seq_1',
        # 'read_struct': 'bc_1+umi_1+seq_1',
        'seq_1': {
            'len': 6,
        },
        'fastq': {
            'fpn': to('data/simu/mclumi/seq_errs/permute_0/seq_err_5.fastq.gz'),
            'trimmed_fpn': to('data/simu/mclumi/seq_errs/permute_0/seq_err_5_trimmed.fastq.gz'),
        },
    }
    print(template(params=params))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.deepconvcvae.Simulate import simulate
from umiche.deepconvcvae.Train import train, data_imputation


def data_in(
        params,
        image_size,
):
    return data_imputation(
        image_size=image_size,
        params=params,
    )


def train_deepconvcvae(
        x_train,
        y_train,
        x_test,
        y_test,
        image_size,
        sv_fpn,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        strides=2,
        epochs=1,
):
    return train(
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        image_size=image_size,
        batch_size=batch_size,
        kernel_size=kernel_size,
        filters=filters,
        latent_dim=latent_dim,
        strides=strides,
        epochs=epochs,
        sv_fpn=sv_fpn,
    )


def simu_deepconvcvae(
        image_size,
        batch_size,
        kernel_size,
        filters,
        latent_dim,
        strides,
        epochs,
        num_dots,
        model_fpn,
        sv_fpn,
):
    return simulate(
        image_size=image_size,
        batch_size=batch_size,
        kernel_size=kernel_size,
        filters=filters,
        num_dots=num_dots,
        latent_dim=latent_dim,
        strides=strides,
        epochs=epochs,
        model_fpn=model_fpn,
        sv_fpn=sv_fpn,
    )


if __name__ == '__main__':
    from umiche.path import to

    train_deepconvcvae()

    # print(simu_deepconvcvae(
    #     image_size=144,
    #     batch_size=32,
    #     kernel_size=3,
    #     filters=16,
    #     latent_dim=2,
    #     strides=2,
    #     epochs=1,
    #     num_dots=100,
    #     model_fpn='../data/deepconvcvae/cvae_model50/cvae_model50.tf',
    #     sv_fpn='../data/deepconvcvae/cvae_model50/cvae_model50.h5',
    # ))

    x_train, y_train, x_test, y_test = data_in(
        params={
            'data_fpn': to('data/deepconvcvae/flt.h5'),
            'data_rv_zero_fpn': to('data/deepconvcvae/flt_rv_zero_col.h5'),
        },
        image_size=144,
    )

    print(train_deepconvcvae(
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        image_size=144,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        strides=2,
        epochs=1,
        sv_fpn='./cvae',
    ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from umiche.util.Console import Console


class Visit:

    def __init__(
            self,
            graph,
            verbose=False,
    ):
        self.graph = graph

        self.console = Console()
        self.console.verbose = verbose

    def recursive(
            self,
            node,
    ):
        visited = []
        def recur(node):
            visited.append(node)
            print(visited)
            for neighbor in self.graph[node]:
                print(neighbor)
                if neighbor not in visited:
                    recur(neighbor)
        recur(node)
        return visited

    def d(
            self,
            node,
            node_val_sorted,
            node_set_remaining,
    ):
        visited = set()
        def recur(node):
            visited.add(node)
            # print(visited)
            for neighbor in self.graph[node]:
                # print(neighbor)
                if neighbor not in visited:
                    if neighbor in node_set_remaining:
                        if node_val_sorted[node] >= 2 * node_val_sorted[neighbor] - 1:
                            recur(neighbor)
        recur(node)
        return visited


if __name__ == "__main__":
    graph = {
        'A': ['B', 'C'],
        'B': ['A', 'D'],
        'C': ['A'],
        'D': ['B', 'E', 'F'],
        'E': ['D'],
        'F': ['D'],
    }

    graph_umi_tools = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D'],
        'F': ['D'],
    }

    node_val_sorted = {
        'A': 456,
        'E': 90,
        'D': 72,
        'B': 2,
        'C': 2,
        'F': 1,
    }

    p = Visit(graph)
    # print(p.recursive('A'))

    nodes = [*node_val_sorted.keys()]
    print(nodes)
    node_cp = nodes.copy()
    node_set_remaining = set(node_cp)
    cc_sub = []
    while nodes:
        e = nodes.pop(0)
        if e in node_set_remaining:
            seen = p.d(e, node_val_sorted, node_set_remaining)
            cc_sub.append(list(seen))
            node_set_remaining = node_set_remaining - seen
            print('remaining: {}'.format(node_set_remaining))
        else:
            continue
    print(cc_sub)


import time
import numpy as np


class Extreme:

    def __init__(self, ):
        pass

    def pervector(self, gc_mat, axis=1):
        d_sum = np.sum(gc_mat, axis=axis)
        if axis == 1:
            return gc_mat / d_sum[:, np.newaxis]
        elif axis == 0:
            return (gc_mat.T / d_sum[:, np.newaxis]).T

    def geneIds(self, gc_mat, oriented='maximal', n_top=10, scheme={'method': 'mean',}, is_norm=False):
        """
        ..  Summary:
            --------
            find ids of extreme (highly-expressed or lowly-expressed) genes.

        ..  Description:
            ------------
            gene_ids is a list containing gene ids

        :param gc_mat:
        :param oriented: maximal or minimal
        :param n_top: the number of top-ranked maximal or minimal values
        :param scheme:
        :param is_norm:
        :return:
        """
        print('======>start to find extreme gene ids...')
        if is_norm:
            gc_mat = self.pervector(gc_mat)
        if scheme['method'] == 'mean':
            mean_stime = time.time()
            d = gc_mat.mean(axis=0)
            # print(d)
            print('======>time: {:.2f}s'.format(time.time() - mean_stime))
            return (-d).argsort()[:n_top].tolist() if oriented == 'maximal' else d.argsort()[:n_top].tolist()
        elif scheme['method'] == 'cell_ratio':
            cr_stime = time.time()
            import pandas as pd
            num_cells = gc_mat.shape[0]
            # print(gc_mat.argmax(axis=1))
            t = pd.Series(gc_mat.argmax(axis=1)).value_counts()
            c = t.rename_axis('unique_values').reset_index(name='counts')
            print('======>time for finding extreme gene ids: {:.2f}s'.format(time.time() - cr_stime))
            return c.loc[c.counts > num_cells * scheme['cell_ratio']].unique_values.values.tolist()

    def perCellClsGeneIds(self, gc_mat, cls_vec={}, oriented='maximal', n_top=10, scheme={'method': 'mean',}, is_norm=False):
        """
        ..  Summary:
            --------
            find ids of extreme (highly-expressed or lowly-expressed) genes in clusters.

        ..  Description:
            ------------
            gene_ids which is a python dictionary, is subject to the data structure below:
            {
                {cluster_1: gene_id_list_1},
                {cluster_2: gene_id_list_2},
                {cluster_3: gene_id_list_3},
                ...
                {cluster_n: gene_id_list_n},
            }

        :param gc_mat:
        :param cls_vec:
        :param oriented:
        :param n_top:
        :param scheme:
        :param is_norm:
        :return:
        """
        print('===>start to find extreme gene ids in clusters...')
        if not isinstance(cls_vec, dict):
            raise ValueError
        res = {}
        for k, cls_vec in cls_vec.items():
            cls_stime = time.time()
            # print(self.gc_mat[cls_vec, :])
            res[k] = self.geneIds(gc_mat[cls_vec, :], oriented=oriented, n_top=n_top, scheme=scheme, is_norm=is_norm)
            print('===>time: {}: {:.2f}s'.format(k, time.time() - cls_stime))
        return res


if __name__ == "__main__":
    gc_mat = np.array([
        [3., 3., 3., 6., 6.],
        [1., 1., 1., 2., 2.],
        [1., 22., 1., 2., 2.]
    ])
    p = Extreme()
    print(p.geneIds(gc_mat, oriented='maximal', n_top=3, scheme={'method': 'cell_ratio', 'cell_ratio': 0.2}))

    cls_vec = {0: [0, 2], 1: [1]}
    print(p.perCellClsGeneIds(
        gc_mat,
        cls_vec,
        oriented='maximal', n_top=3, scheme={'method': 'cell_ratio', 'cell_ratio': 0.2}
    ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from functools import wraps


class Writer:

    def __call__(self, deal):
        generic = self.generic
        excel = self.excel
        @wraps(deal)
        def write(ph, *args, **kwargs):
            res = deal(ph, **kwargs)
            keys = [*kwargs.keys()]
            if kwargs['type'] == 'generic':
                generic(
                    df=kwargs['df'],
                    sv_fpn=kwargs['sv_fpn'],
                    df_sep='\t' if 'df_sep' not in keys else kwargs['df_sep'],
                    id_from=0 if 'id_from' not in keys else kwargs['id_from'],
                    header=None if 'header' not in keys else kwargs['header'],
                    index=False if 'index' not in keys else kwargs['index'],
                )
            elif kwargs['type'] == 'excel':
                excel(
                    df=kwargs['df'],
                    sv_fpn=kwargs['sv_fpn'],
                    sheet_name='Sheet1' if 'sheet_name' not in keys else kwargs['sheet_name'],
                    id_from=0 if 'id_from' not in keys else kwargs['id_from'],
                    header=None if 'header' not in keys else kwargs['header'],
                    index=False if 'index' not in keys else kwargs['index'],
                )
            return res
        return write

    def generic(self, df, sv_fpn, df_sep='\t', header=None, index=False, id_from=0):
        """

        Parameters
        ----------
        df
        sv_fpn
        df_sep
        header
        index
        id_from

        Returns
        -------

        """
        df_ = pd.DataFrame(df)
        # df_.index = df_.index + id_from
        return df_.to_csv(
            sv_fpn,
            sep=df_sep,
            header=header,
            index=index
        )

    def excel(self, df, sv_fpn=None, sheet_name='Sheet1', header=None, index=False, id_from=0):
        """

        Parameters
        ----------
        df
        sv_fpn
        sheet_name
        header
        index
        id_from

        Returns
        -------

        """
        df_ = pd.DataFrame(df)
        df_.index = df_.index + id_from
        return df_.to_excel(
            sv_fpn,
            sheet_name=sheet_name,
            header=header,
            index=index
        )


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd

from umiche.util.Hamming import Hamming
from umiche.util.Console import Console


class Gadgetry:

    def __init__(
            self,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def num_removed_uniq_umis(
            self,
            df_row,
            by_col,
    ):
        """
        It calculates the number of the removed unique UMIs, observed on a given genomic position,
        after CC, Adjacency, Directional, MCL, MCL_val, or MCL_ed.

        Parameters
        ----------
        df_row
            object - a pandas-like df row
        by_col
            str - a column name in question

        Returns
        -------
            int - the sum of deduplicated unique UMI counts per position

        """
        return df_row['num_uniq_umis'] - len(df_row[by_col])

    def num_removed_reads(
            self,
            df_row,
            by_col,
    ):
        """
        It calculates the number of the removed reads in total.

        Parameters
        ----------
        df_row
            object - a pandas-like df row
        by_col
            str - a column name in question

        Returns
        -------
            int - the total counts of deduplicated reads per position

        """
        ### @@ set(df_row['uniq_repr_nodes'])
        # {0, 1, 2, ..., 1948}
        ### @@ len(set(df_row['uniq_repr_nodes']))
        # 1949
        ### @@ set(df_row[by_col])
        # {2}
        ### @@ len(set(df_row[by_col]))
        # 1
        diff_nodes = set(df_row['uniq_repr_nodes']) - set(df_row[by_col])
        ### @@ len(diff_nodes)
        # 1948

        if diff_nodes != set():
            # print(diff_nodes)
            umi_val_cnt_dict = df_row['vignette']['df_umi_uniq_val_cnt'].to_dict()
            ### @@ df_row['vignette']['df_umi_uniq_val_cnt'].to_dict()
            # {2: 55, 780: 51, 416: 47, ..., 1948: 1}
            return sum(umi_val_cnt_dict[node] for node in diff_nodes)
        else:
            return 0

    def umimax(
            self,
            df_row,
            by_col,
    ):
        """
        It returns ids of UMIs (i.e., representatives in their groupped UMIs) that has the
        highest count among all reads in their given genomic position.

        Examples
        --------
        {0: [0, 77, 81, ..., 1016], 1: [42, 46, 12], ..., 100: [2, 3, 5]}
        if no. 77 UMI has the highest count among all reads in 0, it will be added
        to umi_maxval_ids.

        Parameters
        ----------
        df_row
            a row of a pandas dataframe
        by_col
            a name of a pandas dataframe column

        Returns
        -------
        a list

        """
        umi_val_cnts = df_row['vignette']['df_umi_uniq_val_cnt']
        ### @@ umi_val_cnts
        # 2       55
        # 780     51
        # 416     47
        #         ..
        # 1948     1
        # Name: count, Length: 1949, dtype: int64
        umi_maxval_ids = []
        for k_c, nodes in df_row[by_col].items():
            self.console.print('key: {} nodes: {}'.format(k_c, nodes))
            ### @@ k_c, nodes
            # 0 [0, 77, 81, ..., 1016]
            ### @@ umi_val_cnts.loc[umi_val_cnts.index.isin(nodes)]
            # 2       55
            # 780     51
            # 416     47
            #         ..
            # 1948     1
            # Name: count, Length: 1949, dtype: int64
            ### @@ umi_val_cnts.loc[umi_val_cnts.index.isin(nodes)].idxmax()
            # 2
            ### @@ umi_val_cnts.loc[umi_val_cnts.index.isin(nodes)].max()
            # 55
            umi_max = umi_val_cnts.loc[umi_val_cnts.index.isin(nodes)].idxmax()
            umi_maxval_ids.append(umi_max)
        return umi_maxval_ids

    def length(
            self,
            df_val,
    ):
        """

        Parameters
        ----------
        df_val
            a pandas dataframe row

        Returns
        -------
            int - the length of the list

        """
        return len(df_val)

    def decompose(
            self,
            list_nd,
    ):
        """

        Parameters
        ----------
        x

        Returns
        -------

        """
        # print(list_nd)
        # print(len(list_nd))
        list_md = []
        for i in list_nd:
            list_md = list_md + i
        self.console.print('======># of the total reads left after deduplication: {}'.format(len(list_md)))
        # print(len(list_md))
        return list_md

    def bamids(
            self,
            df_row,
            by_col,
    ) -> list:
        """
        It outputs bamids of UMIs that are representative of all nodes in each group.

        Parameters
        ----------
        df_row
            a row of a pandas dataframe
        by_col
            a name of a pandas dataframe column

        Returns
        -------

        """
        uniq_umi_id_to_bam_id_dict = df_row['vignette']['uniq_umi_id_to_bam_id_dict']
        ### @@ uniq_umi_id_to_bam_id_dict
        # {0: 0, 1: 1, 2: 2, ..., 1948: 20609}
        ### @@ len(uniq_umi_id_to_bam_id_dict)
        # 1949
        list_1d = df_row[by_col]
        ### @@ list_1d
        # [2, 780, 416, ..., 1761]
        return [uniq_umi_id_to_bam_id_dict[node] for node in list_1d]

    def ed_ave(
            self,
            df_row,
            by_col,
    ):
        repr_nodes = df_row[by_col]
        # print(repr_nodes)
        node_len = len(repr_nodes)
        int_to_umi_dict = df_row['vignette']['int_to_umi_dict']
        if node_len != 1:
            ed_list = []
            for i in range(node_len):
                for j in range(i + 1, node_len):
                    ed_list.append(Hamming().general(
                        s1=int_to_umi_dict[repr_nodes[i]],
                        s2=int_to_umi_dict[repr_nodes[j]],
                    ))
            # print(pd.Series(ed_list).value_counts())
            return np.ceil(sum(ed_list) / (len(ed_list)))
        else:
            return -1

    def eds_(
            self,
            df_row,
            by_col,
    ):
        """"""
        # print(df_row.index)
        repr_nodes = df_row[by_col]
        int_to_umi_dict = df_row['vignette']['int_to_umi_dict']
        umi_val_cnts = df_row['vignette']['df_umi_uniq_val_cnt']
        # print(repr_nodes)
        # if len(repr_nodes) == len(np.unique(repr_nodes)):
        #     print(True)
        # else:
        #     print(False)
        node_len = len(repr_nodes)
        if node_len != 1:
            ed_list = []
            for i in range(node_len):
                for j in range(i + 1, node_len):
                    if repr_nodes[i] != repr_nodes[j]:
                        ed_list = ed_list + [Hamming().general(
                            int_to_umi_dict[repr_nodes[i]],
                            int_to_umi_dict[repr_nodes[j]])
                        ] * (umi_val_cnts.loc[repr_nodes[i]] * umi_val_cnts.loc[repr_nodes[j]])
            return round(sum(ed_list) / len(ed_list))
        else:
            return -1


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from umiche.util.Reader import Reader as freader
from umiche.deduplicate.pipeline import Config


class protSC(Config.config):

    def __init__(self, fpns):
        """
        self.gmat, _, _ = fromSimulator(simulator='SPsimSeqFixSM').run()
            self.cell_map = {k: v for k, v in enumerate(self.gmat.columns)}
            self.gene_map = {k: v for k, v in enumerate(self.gmat.index)}
            # print(self.cell_map)
            # print(self.gene_map)
            csr_ = coo_matrix(self.gmat)
            print(csr_)
            self.gbyc_arr = np.transpose([
                csr_.row.tolist(),
                csr_.col.tolist(),
                csr_.data.tolist(),
            ]).astype(np.int)
            print(pd.DataFrame(self.gbyc_arr, columns=['cell', 'gene', 'gc_cnt']).to_csv(
                to('data/simu/monomer/sc/seq_errs/ground_truth.txt'), header=True, index=False, sep='\t'))
        :param fpns:
        """
        super(protSC, self).__init__()
        self.freader = freader()
        self.df_gt = self.freader.generic(df_fpn=fpns['ground_truth'], header=0)[:]
        print(self.df_gt['gc_cnt'].sum())
        self.df_gt['gc_tag'] = self.df_gt.apply(lambda x: 'cell'+str(x[0] + 1) + ' ' + 'gene' + str(x[1] + 1), axis=1)
        print(self.df_gt)
        self.df = pd.DataFrame()
        df_melt = pd.DataFrame()
        for method, fpn in fpns['methods'].items():
            df_met = self.freader.generic(df_fpn=fpn, header=0)[:]
            # print(df_met)
            df_met = df_met.sub(self.df_gt['gc_cnt'], axis='rows')
            df_met = df_met.div(self.df_gt['gc_cnt'], axis='rows')
            print(df_met.div(self.df_gt['gc_cnt'], axis='rows'))

            # df_met = (df_met - 50) / 50
            df_met['gc_tag'] = self.df_gt['gc_tag']
            metric_map = {'metric' + str(k): v for k, v in enumerate(self.seq_errs)}
            print(metric_map)
            df_met_melt = pd.melt(df_met, 'gc_tag', var_name='seq_err')
            df_met_melt_ = pd.DataFrame(df_met_melt, columns = ['gc_tag', 'seq_err', 'value'])
            df_met_melt_['method'] = method
            df_met_melt_['se_val'] = df_met_melt_['seq_err'].apply(lambda x: metric_map[x])
            print(df_met_melt_)
            df_met['method'] = method

            self.df = pd.concat([self.df, df_met], axis=0)
            df_melt = pd.concat([df_melt, df_met_melt_], axis=0)
        print(self.df)
        self.df_direc = self.df[self.df['method'] == 'directional']
        self.df_mcl_val = self.df[self.df['method'] == 'mcl_val']
        self.df_mcl_ed = self.df[self.df['method'] == 'mcl_ed']
        print(self.df_direc.iloc[1].values.shape)
        # grid = sns.FacetGrid(df_melt, col="gc_tag", hue="method", palette=['r', 'b', 'g'],
        #                      col_wrap=4, height=1.5)
        #
        # # Draw a horizontal line to show the starting point
        # # grid.refline(y=0, linestyle=":")
        #
        # # Draw a line plot to show the trajectory of each random walk
        # grid.map(plt.plot, "se_val", "value", marker="o")
        #
        # # Adjust the tick positions and labels
        # # grid.set(xticks=np.arange(5), yticks=[-3, 3],
        # #          xlim=(-.5, 4.5), ylim=(-3.5, 3.5))
        #
        # # Adjust the arrangement of the plots
        # grid.fig.tight_layout(w_pad=1)
        # plt.show()

    def separateline(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        row_num = 5
        col_num = 8
        fig, ax = plt.subplots(nrows=row_num, ncols=col_num, figsize=(12, 7), sharey=False, sharex='all')
        c = 0
        for i_row in range(row_num):
            for i_col in range(col_num):
                print('asdsad', len( self.seq_errs))
                ax[i_row][i_col].plot(
                    self.seq_errs,
                    self.df_direc.iloc[c].values[:15],
                    color='tab:green',
                    alpha=0.9,
                    label='directional',
                )
                ax[i_row][i_col].plot(
                    self.seq_errs,
                    self.df_mcl_ed.iloc[c].values[:15],
                    color='tab:blue',
                    alpha=0.9,
                    label='mcl_ed',
                )
                ax[i_row][i_col].plot(
                    self.seq_errs,
                    self.df_mcl_val.iloc[c].values[:15],
                    color='crimson',
                    alpha=0.9,
                    label='mcl_val',
                )
                ax[i_row][i_col].spines['right'].set_visible(False)
                ax[i_row][i_col].spines['top'].set_visible(False)
                ax[i_row][i_col].set_title(self.df_mcl_val.iloc[c].values[15] + '\ncnt:' + str(self.df_gt.iloc[c].values[2]), fontsize=8)
                # print(self.seq_errs[::2])

                ylim_low, ylim_up = ax[i_row][i_col].get_ylim()
                # ax[i_row][i_col].set_ylim(ylim_low, ylim_up + 0.01)
                # print(ylim_low)
                ylims = [round(i) for i in np.linspace(0, ylim_up, 3)]
                ax[i_row][i_col].set_yticks(ylims)
                ax[i_row][i_col].set_yticklabels(ylims, fontsize=8)

                xlims = np.linspace(self.seq_errs[0], self.seq_errs[-1], 4)
                ax[i_row][i_col].set_xticks(xlims)
                ax[i_row][i_col].set_xticklabels(['{:.1e}'.format(i) for i in xlims], rotation=20, fontsize=8)
                # ax[i_row][i_col].legend(loc='upper left', fontsize=5)
                c += 1

        fig.add_subplot(111, frameon=False)
        plt.tick_params(labelcolor='none', which='both', top=False, bottom=False, left=False, right=False)
        # plt.xlabel('Sequencing error rate', fontsize=12)
        plt.ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)

        fig.supxlabel('Sequencing error rate', fontsize=12)
        # fig.supylabel(r'$\frac{N_e-N_t}{N_t}$')
        # plt.tight_layout()
        fig.subplots_adjust(
            top=0.88,
            bottom=0.10,
            left=0.08,
            right=0.98,
            hspace=0.60,
            wspace=0.30
        )

        lines = []
        labels = []
        axLine, axLabel = ax[0][0].get_legend_handles_labels()
        lines.extend(axLine)
        labels.extend(axLabel)
        plt.legend(lines, labels, fontsize=9, bbox_to_anchor=(0.5, 1.15), loc=9, ncol=7)
        plt.show()
        return


if __name__ == "__main__":
    from umiche.path import to
    DEFINE = {
        'fpns': {
            'methods': {
                'directional': to('data/simu/sc/seq_errs/directional.txt'),
                'mcl_ed': to('data/simu/sc/seq_errs/mcl_ed.txt'),
                'mcl_val': to('data/simu/sc/seq_errs/mcl_val.txt'),
            },
            'ground_truth': to('data/simu/sc/seq_errs/ground_truth.txt'),
        },
    }
    p = protSC(DEFINE['fpns'])

    print(p.separateline())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from umiche.deepconvcvae.h5o.Reader import reader as h5reader
from umiche.deepconvcvae.cv.Summary import summary as cvsum
from umiche.deepconvcvae.train.Summary import summary as trainsum
from umiche.deepconvcvae.boxplot.GCMat import gcmat as splboxplot
from umiche.deepconvcvae.cv.Split import split as cvsplit


class preprocess:

    def __init__(self, data_fpn, train_ratio=3, test_ratio=2):
        self.cvsum = cvsum()
        self.trainsum = trainsum()
        self.splboxplot = splboxplot()
        self.h5reader = h5reader()
        self.cvsplit = cvsplit()
        self.data_dict = self.h5reader.gcmatLit(data_fpn)
        self.cv_dict = self.cvsplit.sss(data_dict=self.data_dict, train_ratio=train_ratio, test_ratio=test_ratio)
        # print(self.cv_dict)
        print([*self.data_dict['dict_cell_cls'].keys()])
        # print(len(self.data_dict['dict_genes']))
        # print(len(self.data_dict['dict_cells']))

    def plotSummary(self, ):
        # ### /* train sum plot */
        train_extreme_dict = self.trainsum.extreme(
            gc_mat=self.data_dict['gc_mat'],
            cls_vec=self.data_dict['dict_cell_cls']
        )
        train_std_dict = {'real': {'val_dict_arr_1d': train_extreme_dict}}
        print('asdasd',train_std_dict)
        self.splboxplot.boxplot(train_std_dict)

        # ### /* cv sum plot */
        cv_std_dict = self.cvsum.extreme(cv_dict=self.cv_dict, gc_mat=self.data_dict['gc_mat'], type='train')
        self.splboxplot.boxplot(cv_std_dict)

    def recipeData(self, cv=0):
        df_cell_cls = pd.DataFrame.from_dict(self.data_dict['dict_cells'], orient='index', columns=['cluster'])
        uniq = df_cell_cls.cluster.unique()
        clusters = {e: i for i, e in enumerate(uniq)}
        df_cell_cls['cls'] = df_cell_cls.cluster.apply(lambda x: clusters[x])
        # print(df_cell_cls)

        train_retrv_ids = self.cv_dict['cv' + str(cv)]['train_index']
        test_retrv_ids = self.cv_dict['cv' + str(cv)]['test_index']

        # from ext.dacube.dataset.gcmat.preprocess.Remove import remove as gcrv
        # self.data_dict['gc_mat'] = gcrv().lowlyexpressed(self.data_dict['gc_mat'], thres=1000)

        x_train = self.data_dict['gc_mat'][train_retrv_ids, :]
        y_train = df_cell_cls['cls'].values[train_retrv_ids]
        # print(x_train.shape)
        # print(y_train.shape)

        x_test = self.data_dict['gc_mat'][test_retrv_ids, :]
        y_test = df_cell_cls['cls'].values[test_retrv_ids]
        # print(x_test.shape)
        # print(y_test.shape)
        del self.data_dict
        del self.cv_dict
        return (x_train, y_train), (x_test, y_test)

    def augment(self, x_train, y_train, x_test, y_test, image_size):
        import numpy as np
        # print(x_train.shape)
        # print(y_train.shape)
        x_train_height = x_train.shape[0]
        x_test_height = x_test.shape[0]
        x_train_width = x_train.shape[1]

        img_col_offset_size = image_size * image_size - x_train_width
        train_col_offset_mat = np.zeros([x_train_height, img_col_offset_size])
        test_col_offset_mat = np.zeros([x_test_height, img_col_offset_size])
        x_train = np.concatenate([x_train, train_col_offset_mat], axis=1)
        x_test = np.concatenate([x_test, test_col_offset_mat], axis=1)
        x_train = np.reshape(x_train, newshape=(-1, image_size, image_size))
        x_test = np.reshape(x_test, newshape=(-1, image_size, image_size))
        # print(x_train)
        # print(x_test)

        x_train = np.reshape(x_train, [-1, image_size, image_size, 1])
        x_test = np.reshape(x_test, [-1, image_size, image_size, 1])
        print(x_train.shape)
        print(x_test.shape)
        std_max_mark = np.maximum(np.amax(x_train), np.amax(x_test))
        print(std_max_mark)
        x_train = x_train.astype('float32') / std_max_mark
        x_test = x_test.astype('float32') / std_max_mark
        print(y_train)
        print(y_test)
        return (x_train, y_train), (x_test, y_test)


if __name__ == "__main__":
    from umiche.path import to

    params = {
        'data_fpn': to('data/sc/gmat/pbmc68k/dA_flt/flt.h5'),
        'data_rv_zero_fpn': to('data/sc/gmat/pbmc68k/dA_flt/flt_rv_zero_col.h5'),
    }
    p = preprocess(params['data_rv_zero_fpn'])
    p.plotSummary()
    # (x_train, y_train), (x_test, y_test) = p.recipeData(cv=0)
    # print((x_train.shape, y_train.shape), (x_test.shape, y_test.shape))
    # print(len(p.data_dict))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import List, Dict

from collections import deque
from umiche.util.Console import Console


class CC:

    def __init__(
            self,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose


    def deque(
            self,
            graph : Dict,
    ):
        """

        Parameters
        ----------
        graph
            1d dict of an adjacency list

        Returns
        -------
        2d list, each 1d list representing a component

        """
        visited = set()
        for root, nbrs in graph.items():
            if root not in visited:
                visited.add(root)
                component = []
                queue = deque([root])
                self.console.print('======> root {} has not been visited'.format(root))
                self.console.print('======> a queue built by root {} is {}'.format(root, queue))
                while queue:
                    self.console.print('=========> a queue built by each root node {}'.format(queue))
                    node = queue.popleft()
                    self.console.print('=========> node: {}'.format(node))
                    component.append(node)
                    for nbr in graph[node]:
                        if nbr not in visited:
                            visited.add(nbr)
                            queue.append(nbr)
                self.console.print('=========> visited nodes {}'.format(visited))
                yield component
            else:
                self.console.print('=========> root {} has been visited'.format(root))
                continue

    def set(
            self,
            graph : Dict,
    ) -> List:
        """
        Examples
        --------
        >>>seen = set()
        >>>def component(node):
        >>>    nodes = set([node])
        >>>    while nodes:
        >>>        node = nodes.pop()
        >>>        seen.add(node)
        >>>        nodes |= neighbors[node] - seen
        >>>        yield node
        >>>    for node in neighbors:
        >>>        if node not in seen:
        >>>            yield component(node)

        Parameters
        ----------
        graph
            1d dict of an adjacency list

        Returns
        -------
        2d list, each 1d list representing a component

        """
        visited = set()
        components = []
        for root, nbrs in graph.items():
            if root not in visited:
                visited.add(root)
                component = []
                queue = [root]
                self.console.print('======> root {} has not been visited'.format(root))
                self.console.print('======> a queue built by root {} is {}'.format(root, queue))
                while queue:
                    self.console.print('======> a queue built by each root node {}'.format(queue))
                    node = queue.pop(0)
                    self.console.print('======> node: {}'.format(node))
                    component.append(node)
                    for nbr in graph[node]:
                        if nbr not in visited:
                            visited.add(nbr)
                            queue.append(nbr)
                self.console.print('======> visited nodes {}'.format(visited))
                components.append(component)
            else:
                self.console.print('=========>root {} has been visited'.format(root))
        return components


if __name__ == "__main__":
    graph_adj_mclumi = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }
    p = CC()

    print(list(p.deque(graph=graph_adj_mclumi)))

    print(p.set(graph=graph_adj_mclumi))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


class Filter:

    def method(self, ):
        return {
            'single_start': self.singleStart,
            'paired': self.paired,
        }

    def singleStart(self, x, start, end):
        """

        Parameters
        ----------
        x
        start
        end

        Returns
        -------

        """
        return x[start: end]
        # len_dict = {}
        # for key, val in start_len.items():
        #     len_dict[key] = 0
        #     for j in val:
        #         len_dict[key] += self.read_summary[j]['len']
        # print(len_dict)
        # # return x[b_len: b_len+self.read_summary['umi']['len']]
        # # return x[a_len: a_len+self.args['umi']['len']]

    def paired(self, x, rule):
        """

        Parameters
        ----------
        x
        rule

        Returns
        -------

        """
        a_len = 0
        b_len = 0
        for j in rule[0]:
            b_len += self.args[j]['len']
        for j in rule[1]:
            a_len += self.args[j]['len']
        return x[b_len: b_len+self.args['umi']['len']]
        # return x[a_len: a_len+self.args['umi']['len']]


if __name__ == "__main__":
    from umiche.path import to

    DEFINE = {
        'umi': {
            'len': 12,
        },
        # 'seq_struct': 'umi*seq',
        'seq_struct': 'primer*umi*seq*umi*primer',
        'primer': {
            'len': 20,
        },
        'seq': {
            'len': 20,
        },
        'fastq': {
            'path': to('data/'),
            'name': 'simu',
        },
    }
    p = Filter(DEFINE)

    umis = p.cus()

    print()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import time
import h5py
import json


class writer(object):

    def __init__(self, ):
        pass

    def gcmatLit(self, gc_dict, sv_fpn, compression_level=4):
        """
        ..  Description:
            ------------
            create a light version of a gene-by-cell matrix from a gc dict consisting of
            a gc matrix of a csr version (see link 1) with three vectors:
                1) gene expression values;
                2) row indices of cells;
                3) index pointers;
            and with three json encoded dicts of
                4) gene notations and
                5) cell names,
                6) cells clusters (using cell names)
            , respectively.

            gc_mat_vals: data of scipy csr matrix

            gc_mat_ids: indices of scipy csr matrix

            gc_mat_id_pointers: indptr of scipy csr matrix

            gene_notations:
            {
                gene_notation_1: 1, # int value
                gene_notation_2: 2, # int value
                gene_notation_3: 3, # int value
                ...
                gene_notation_n: n, # int value
            }

            cell_names:
            {
                cell_name_1: 1, # int value
                cell_name_2: 2, # int value
                cell_name_3: 3, # int value
                ...
                cell_name_m: m, # int value
            }

            dict_cell_cls (i.e., cell clusters) are structured as
            {
                cell_cluster_1: cell_indices_in_cluster_1, # python list
                cell_cluster_2: cell_indices_in_cluster_2, # python list
                cell_cluster_3: cell_indices_in_cluster_3, # python list
                ...
                cell_cluster_k: cell_indices_in_cluster_k, # python list
            }

        ..  See:
            ----
            1. https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.html;
            2. https://docs.h5py.org/en/stable/high/dataset.html?highlight=compression#lossless-compression-filters

        :param gc_dict:
        :param sv_fpn:
        :param compression_level: corresponds to compression_opts in hdf5 (see link 2).
        :return: a hdf5 ob
        """
        l2d_stime = time.time()
        with h5py.File(sv_fpn, 'w') as f:
            g1 = f.create_group('default')
            g1.create_dataset(
                'gc_mat_vals',
                data=gc_dict['gc_mat_vals'],
                compression='gzip',
                compression_opts=compression_level,
            )
            g1.create_dataset(
                'gc_mat_ids',
                data=gc_dict['gc_mat_ids'],
                compression='gzip',
                compression_opts=compression_level,
            )
            g1.create_dataset(
                'gc_mat_id_pointers',
                data=gc_dict['gc_mat_id_pointers'],
                compression='gzip',
                compression_opts=compression_level,
            )
            g1.create_dataset(
                'dict_genes',
                data=json.dumps(gc_dict['dict_genes'])
            )
            g1.create_dataset(
                'dict_cells', data=json.dumps(gc_dict['dict_cells'])
            )
            g1.create_dataset(
                'dict_cell_cls', data=json.dumps(gc_dict['dict_cell_cls'])
            )
        f.close()
        print('===>time for saving h5: {:.2f}'.format(time.time() - l2d_stime))
        # with h5py.File(sv_fpn, 'r') as f:
        #     print(f['default/gc_mat_vals'][()])
        #     print(json.loads(f['default/cell_names'][()]))
        return 0


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.simu.Parameter import Parameter as params
from umiche.deduplicate.io.Stat import Stat as dedupstat

from umiche.util.Reader import Reader as freader


class Line:

    def __init__(
            self,
            scenarios: Dict,
            methods: Dict,
            umi_gt_cnt: int = 50,
            param_fpn: str = None,
    ):
        self.scenarios = scenarios
        self.methods = methods
        self.umi_gt_cnt = umi_gt_cnt
        self.param_fpn = param_fpn

        self.params = params(param_fpn=self.param_fpn)

        self.freader = freader()

        self.dedupstat = dedupstat(
            scenarios=self.scenarios,
            methods=self.methods,
            param_fpn=self.param_fpn,
        )

        self.df_inflat, self.df_exp = self.dedupstat.df_inflat_exp

        # sns.set(font="Verdana")
        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def draw(
            self,
            num_img_row=1,
            num_img_col=2,
    ):
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(10, 5), sharey=False, sharex=False)
        palette = sns.color_palette("Set2") # Paired
        print(self.df_inflat)
        print(self.df_exp)
        dict_df = {
            'Inflation': self.df_inflat,
            'Expansion': self.df_exp,
        }
        scenarios = self.df_inflat.columns
        for i, (key, df) in enumerate(dict_df.items()):
            for j, scenario in enumerate(scenarios):
                ax[i].plot(
                    df.index.tolist(),
                    df[scenario].values.tolist(),
                    label=scenario,
                    linewidth=2.5,
                    marker='o',
                    markerfacecolor='white',
                    alpha=0.9,
                    color=palette[j],
                )
            ax[i].legend(title='', ncol=1, fontsize=12, )
            ax[i].set_xlabel(key, fontsize=14)
            ax[i].set_ylabel(r'($\frac{N_e-N_t}{N_t}$)', fontsize=18)
            ax[i].spines['right'].set_visible(False)
            ax[i].spines['top'].set_visible(False)
        ax[0].axvline(x=3.5, color='grey', linestyle='dotted')
        ax[1].axvline(x=4, color='grey', linestyle='dotted')

        plt.subplots_adjust(
            top=0.96,
            bottom=0.12,
            left=0.09,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        plt.show()


if __name__ == "__main__":
    from umiche.path import to

    p = Line(
        scenarios={
            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'ampl_rates': 'Amplification rate',
            'umi_lens': 'UMI length',
            'seq_deps': 'Sequencing depth',
        },

        methods={
            'mcl': 'MCL',
        },

        param_fpn=to('data/params.yml'),
    )

    p.draw()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import networkx as nx


class undirected:

    def adaa(self, ):
        import matplotlib.pyplot as plt
        import networkx as nx

        G = nx.Graph()

        G.add_edge('a', 'b', weight=0.6)
        G.add_edge('a', 'c', weight=0.2)
        G.add_edge('c', 'd', weight=0.1)
        G.add_edge('c', 'e', weight=0.7)
        G.add_edge('c', 'f', weight=0.9)
        G.add_edge('a', 'd', weight=0.3)

        elarge = [(u, v) for (u, v, d) in G.edges(data=True) if d['weight'] > 0.5]
        esmall = [(u, v) for (u, v, d) in G.edges(data=True) if d['weight'] <= 0.5]

        pos = nx.spring_layout(G)  # positions for all nodes

        # nodes
        nx.draw_networkx_nodes(G, pos, node_size=700)

        # edges
        nx.draw_networkx_edges(G, pos, edgelist=elarge,
                               width=6)
        nx.draw_networkx_edges(G, pos, edgelist=esmall,
                               width=6, alpha=0.5, edge_color='b', style='dashed')

        # labels
        nx.draw_networkx_labels(G, pos, font_size=20, font_family='sans-serif')

        plt.axis('off')
        plt.show()


    def plot_mute_graph(self, ):
        """Plot MuTE example network.

        Network of 5 AR-processes, which is used as an example the paper
        on the MuTE toolbox (Montalto, PLOS ONE, 2014, eq. 14). The
        network consists of five autoregressive (AR) processes with model
        orders 2 and les and the following (non-linear) couplings:

            >>> 0 -> 1, u = 2
            >>> 0 -> 2, u = 3
            >>> 0 -> 3, u = 2 (non-linear)
            >>> 3 -> 4, u = 1
            >>> 4 -> 3, u = 1

        Returns:
            Figure handle
                Figure object from the matplotlib package
        """
        graph = nx.DiGraph()
        graph.add_nodes_from(np.arange(5))
        # graph.add_edges_from([(0, 1), (0, 2), (0, 3), (3, 4), (4, 3)])
        graph.add_weighted_edges_from([(0, 1, 2), (0, 2, 3), (0, 3, 2), (3, 4, 1),
                                       (4, 3, 1)], weight='delay')
        pos = {
            0: np.array([1, 1]),
            1: np.array([0, 2]),
            2: np.array([0, 0]),
            3: np.array([2, 1]),
            4: np.array([3, 1]),
        }
        fig = plt.figure()
        nx.draw(graph, pos=pos, with_labels=True, node_size=900, alpha=1.0,
                node_color='cadetblue', font_weight='bold',
                edge_color=['r', 'k', 'r', 'k', 'k'])
        # nx.draw_networkx_edge_labels(graph, pos=pos)
        plt.text(2, 0.1, 'non-linear interaction in red')
        # see here for an example on how to plot edge labels:
        # http://stackoverflow.com/questions/10104700/how-to-set-networkx-edge-labels-offset-to-avoid-label-overlap
        plt.show()
        return fig


if __name__ == "__main__":
    p = undirected()
    print(p.adaa())
    print(p.plot_mute_graph())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import pandas as pd
from umiche.simu.Parameter import Parameter as params

from umiche.util.Reader import Reader as freader
from umiche.util.Console import Console


class Stat:

    def __init__(
            self,
            scenarios: Dict,
            methods: Dict,
            umi_gt_cnt: int = 50,
            param_fpn : str = None,
            is_trans : bool = True,
            verbose: bool = True,
    ):
        self.scenarios = scenarios
        self.methods = methods
        self.umi_gt_cnt = umi_gt_cnt
        self.freader = freader()
        self.params = params(param_fpn=param_fpn)
        self.is_trans = is_trans

        self.console = Console()
        self.console.verbose = verbose

    @property
    def df_dedup(self, ):
        df = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            for method, method_formal in self.methods.items():
                self.console.print("=========>method: {}".format(method_formal))
                ph = '/' if method_formal != 'Birch' else '/birch/'
                df_sce_met = self.freader.generic(
                    # df_fpn=self.params.work_dir + scenario + '/' + method + '_dedup.txt',
                    df_fpn=self.params.work_dir + scenario + ph + method + '_dedup.txt',
                    header=0,
                )
                if self.is_trans:
                    df_sce_met = (df_sce_met - self.umi_gt_cnt) / self.umi_gt_cnt
                df_sce_met['mean'] = df_sce_met.mean(axis=1)
                df_sce_met['max'] = df_sce_met.max(axis=1)
                df_sce_met['min'] = df_sce_met.min(axis=1)
                df_sce_met['std'] = df_sce_met.std(axis=1)
                df_sce_met['mean-min'] = df_sce_met['std']
                df_sce_met['max-mean'] = df_sce_met['std']
                df_sce_met['scenario'] = scenario_formal
                df_sce_met['method'] = method_formal
                df_sce_met['metric'] = [str(x) for x in self.params.varied[scenario]]
                df = pd.concat([df, df_sce_met], axis=0)
        return df

    @property
    def df_runtime(self, ):
        df = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            for method, method_formal in self.methods.items():
                self.console.print("=========>method: {}".format(method_formal))
                # print(self.params.work_dir + scenario + '/' + method + '.txt')
                df_sce_met = self.freader.generic(
                    df_fpn=self.params.work_dir + scenario + '/' + method + '.txt',
                    header=0,
                )
                # print(df_sce_met)
                if self.is_trans:
                    df_sce_met = (df_sce_met - self.umi_gt_cnt) / self.umi_gt_cnt
                df_sce_met['mean'] = df_sce_met.mean(axis=1)
                # print('1', df_sce_met)
                # print('asdsad', df_sce_met['mean'])
                df_sce_met['max'] = df_sce_met.max(axis=1)
                df_sce_met['min'] = df_sce_met.min(axis=1)
                df_sce_met['std'] = df_sce_met.std(axis=1)
                df_sce_met['mean-min'] = df_sce_met['std']
                df_sce_met['max-mean'] = df_sce_met['std']
                df_sce_met['scenario'] = scenario_formal
                df_sce_met['method'] = method_formal
                df_sce_met['metric'] = [str(x) for x in self.params.varied[scenario]]
                df = pd.concat([df, df_sce_met], axis=0)
        return df

    @property
    def df_dedup_set_cover_len(self, ):
        df = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            for method, method_formal in self.methods.items():
                self.console.print("=========>method: {}".format(method_formal))
                df_sce_met = self.freader.generic(
                    df_fpn=self.params.work_dir + scenario + '/' + method + '_dedup.txt',
                    header=0,
                )
                if self.is_trans:
                    df_sce_met = df_sce_met.applymap(lambda x: sum([int(x) for x in x.split(';')]))
                    # print(df_sce_met)
                    df_sce_met['mean'] = df_sce_met.mean(axis=1)
                    df_sce_met['max'] = df_sce_met.max(axis=1)
                    df_sce_met['min'] = df_sce_met.min(axis=1)
                    df_sce_met['std'] = df_sce_met.std(axis=1)
                    df_sce_met['mean-min'] = df_sce_met['std']
                    df_sce_met['max-mean'] = df_sce_met['std']
                df_sce_met['scenario'] = scenario_formal
                df_sce_met['method'] = method_formal
                df_sce_met['metric'] = [str(x) for x in self.params.varied[scenario]]
                df = pd.concat([df, df_sce_met], axis=0)
        return df

    @property
    def df_dedup_perm_melt(self, ):
        df = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            for method, method_formal in self.methods.items():
                self.console.print("=========>method: {}".format(method_formal))
                df_sce_met = self.freader.generic(
                    df_fpn=self.params.work_dir + scenario + '/' + method + '_dedup.txt',
                    header=0,
                )
                df_sce_met_T = df_sce_met.T
                df_sce_met_T = (df_sce_met_T - self.umi_gt_cnt) / self.umi_gt_cnt
                df_sce_met_T.columns = [str(x) for x in self.params.varied[scenario]]
                df_sce_met_T['method'] = method_formal
                df = pd.concat([df, df_sce_met_T], axis=0)
        df_melt = pd.melt(df, 'method', var_name="Sequencing error")
        return df_melt

    @property
    def df_dedup_melt(self, ):
        df = pd.melt(
            frame=self.df_dedup[['method', 'metric', 'scenario', 'mean']],
            id_vars=['method', 'metric', 'scenario'],
            # value_vars=['mean',],
        )
        return df

    @property
    def df_trace_cnt(self, ):
        df_apv = pd.DataFrame()
        df_disapv = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            for method, method_formal in self.methods.items():
                self.console.print("=========>method: {}".format(method_formal))
                df_sce_met_apv_perm = self.freader.generic(
                    df_fpn=self.params.work_dir + scenario + '/' + method + '_apv_cnt.txt',
                    header=0,
                )
                df_sce_met_disapv_perm = self.freader.generic(
                    df_fpn=self.params.work_dir + scenario + '/' + method + '_disapv_cnt.txt',
                    header=0,
                )
                df_sce_met_apv_perm = df_sce_met_apv_perm.rename(columns={'scenario': 'metric'})
                df_sce_met_disapv_perm = df_sce_met_disapv_perm.rename(columns={'scenario': 'metric'})

                df_sce_met_apv = df_sce_met_apv_perm.groupby(by=['metric']).agg({'diff_origin': 'mean', 'same_origin': 'mean'}).reset_index()
                df_sce_met_apv['diff_origin_max'] = df_sce_met_apv_perm.groupby(by=['metric']).agg({'diff_origin': 'max'})['diff_origin'].values
                df_sce_met_apv['diff_origin_min'] = df_sce_met_apv_perm.groupby(by=['metric']).agg({'diff_origin': 'min'})['diff_origin'].values
                df_sce_met_apv['same_origin_max'] = df_sce_met_apv_perm.groupby(by=['metric']).agg({'same_origin': 'max'})['same_origin'].values
                df_sce_met_apv['same_origin_min'] = df_sce_met_apv_perm.groupby(by=['metric']).agg({'same_origin': 'min'})['same_origin'].values
                df_sce_met_apv['scenario'] = scenario_formal
                df_sce_met_apv['method'] = method_formal

                df_sce_met_disapv = df_sce_met_disapv_perm.groupby(by=['metric']).agg({'diff_origin': 'mean', 'same_origin': 'mean'}).reset_index()
                df_sce_met_disapv['diff_origin_max'] = df_sce_met_disapv_perm.groupby(by=['metric']).agg({'diff_origin': 'max'})['diff_origin'].values
                df_sce_met_disapv['diff_origin_min'] = df_sce_met_disapv_perm.groupby(by=['metric']).agg({'diff_origin': 'min'})['diff_origin'].values
                df_sce_met_disapv['same_origin_max'] = df_sce_met_disapv_perm.groupby(by=['metric']).agg({'same_origin': 'max'})['same_origin'].values
                df_sce_met_disapv['same_origin_min'] = df_sce_met_disapv_perm.groupby(by=['metric']).agg({'same_origin': 'min'})['same_origin'].values
                df_sce_met_disapv['scenario'] = scenario_formal
                df_sce_met_disapv['method'] = method_formal

                # print(df_sce_met_apv)
                # print(df_sce_met_disapv)

                df_apv = pd.concat([df_apv, df_sce_met_apv], axis=0)
                df_disapv = pd.concat([df_disapv, df_sce_met_disapv], axis=0)
        return {
            "apv": df_apv,
            "disapv": df_disapv,
        }

    @property
    def df_inflat_exp(self, ):
        df_inflat = pd.DataFrame()
        df_exp = pd.DataFrame()
        for scenario, scenario_formal in self.scenarios.items():
            self.console.print("======>scenario: {}".format(scenario_formal))
            df_inflat_sub = self.freader.generic(
                df_fpn=self.params.work_dir + scenario + '/inflat_val.txt',
                header=None,
            )
            df_inflat_sub = df_inflat_sub.rename(columns={0: 'id', 1: 'cnt'})
            df_exp_sub = self.freader.generic(
                df_fpn=self.params.work_dir + scenario + '/exp_val.txt',
                header=None,
            )
            df_exp_sub = df_exp_sub.rename(columns={0: 'id', 1: 'cnt'})
            # print(df_inflat_sub)
            # print(df_exp_sub)
            # print(scenario)
            df_inflat[scenario_formal] = df_inflat_sub['cnt']
            df_exp[scenario_formal] = df_exp_sub['cnt']
        df_inflat = (df_inflat - self.umi_gt_cnt) / self.umi_gt_cnt
        df_exp = (df_exp - self.umi_gt_cnt) / self.umi_gt_cnt
        df_inflat.index = df_inflat_sub['id'].values
        df_exp.index = df_exp_sub['id'].astype(int).values
        return df_inflat, df_exp


if __name__ == "__main__":
    from umiche.path import to

    p = Stat(
        scenarios={
            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'ampl_rates': 'Amplification rate',
            'umi_lens': 'UMI length',
            'seq_deps': 'Sequencing depth',
        },

        methods={
            # 'unique': 'Unique',
            # 'cluster': 'Cluster',
            # 'adjacency': 'Adjacency',
            'directional': 'Directional',
            # 'dbscan_seq_onehot': 'DBSCAN',
            # 'birch_seq_onehot': 'Birch',
            # 'aprop_seq_onehot': 'Affinity Propagation',
            'mcl': 'MCL',
            # 'mcl_val': 'MCL-val',
            # 'mcl_ed': 'MCL-ed',
        },

        param_fpn=to('data/params.yml'),
        verbose=True
    )

    print(p.df_dedup)
    print(p.df_dedup_melt)
    print(p.df_trace_cnt)
    print(p.df_inflat_exp)


# import umi_tools.sam_methods as sam_methods
import umiche.external.UMItoolsSamMethods as sam_methods
import pysam
from umiche.util.Hamming import Hamming
import pandas as pd
import numpy as np
# from umiche.path import to


def edave(x, d):
    repr_nodes = d[x[1]]
    node_len = len(repr_nodes)
    if node_len != 1:
        ed_list = []
        for i in range(node_len):
            for j in range(i + 1, node_len):
                ed_list.append(Hamming().general(
                    s1=repr_nodes[i],
                    s2=repr_nodes[j]
                ))
        return np.ceil(sum(ed_list) / (len(ed_list)))
    else:
        return -1


def convert(options, in_fpn, out_fpn):
    bundle_iterator = sam_methods.get_bundles(
        options,
        metacontig_contig=None,
    )
    infile = pysam.Samfile(in_fpn, 'rb')
    inreads = infile.fetch()
    num_bundled_pos = 0
    umis = []
    nInput = 0
    reads = []
    uniq_umi_cnt = 0
    write_to_bam = pysam.AlignmentFile(out_fpn, "wb", template=infile)
    for i, (bundle, key, status) in enumerate(bundle_iterator(inreads)):
        # print([bundle[umi]["read"] for umi in bundle])
        for j, umi in enumerate(bundle):
            read = bundle[umi]["read"]
            read.set_tag('PO', i)
            for _ in range(bundle[umi]["count"]):
                reads.append(bundle[umi]["read"])
            uniq_umi_cnt += 1
        nInput += sum([bundle[umi]["count"] for umi in bundle])

        umis.append([i.decode('utf-8') for i in bundle.keys()])
        num_bundled_pos += 1

    for y in reads:
        write_to_bam.write(y)
    write_to_bam.close()

    print('# of unique reads {}'.format(len(reads)))
    print('# of bundled pos {}'.format(num_bundled_pos))
    print('# of repeated UMIs in total {}'.format(nInput))
    df = pd.DataFrame(index=np.arange(num_bundled_pos))
    df[1] = np.arange(num_bundled_pos)
    df[2] = df.apply(lambda x: edave(x, umis), axis=1)
    return df[2].value_counts()


# in_fpn = to('example/data/example.bam')
# out_fpn = to('example/data/example_bundle.bam')
# options = {'stats': 'deduplicated',
#            'get_umi_method': 'read_id',
#            'umi_sep': '_',
#            'umi_tag': 'RX',
#            'umi_tag_split': None,
#            'umi_tag_delim': None,
#            'cell_tag': None,
#            'cell_tag_split': '-',
#            'cell_tag_delim': None,
#            'filter_umi': None,
#            'umi_whitelist': None,
#            'umi_whitelist_paired': None,
#            'method': 'directional',
#            'threshold': 1,
#            'spliced': False,
#            'soft_clip_threshold': 4,
#            'read_length': False,
#            'per_gene': False,
#            'gene_tag': None,
#            'assigned_tag': None,
#            'skip_regex': '^(__|Unassigned)',
#            'per_contig': False,
#            'gene_transcript_map': None,
#            'per_cell': False,
#            'whole_contig': False,
#            'detection_method': None,
#            'mapping_quality': 0,
#            'output_unmapped': False,
#            'unmapped_reads': 'discard',
#            'chimeric_pairs': 'use',
#            'unpaired_reads': 'use',
#            'ignore_umi': False,
#            'ignore_tlen': False,
#            'chrom': None,
#            'subset': None,
#            'in_sam': False,
#            'paired': False,
#            'out_sam': False,
#            'no_sort_output': False,
#            'stdin': "<_io.TextIOWrapper name='example.bam' mode='r' encoding='UTF-8'>",
#            'stdlog': "<_io.TextIOWrapper name='<stdout>' mode='w' encoding='UTF-8'>", 'log2stderr': False,
#            'compresslevel': 6,
#            'timeit_file': None,
#            'timeit_name': 'all',
#            'timeit_header': None,
#            'loglevel': 1,
#            'short_help': None,
#            'random_seed': None
#            }
# convert(options=options, in_fpn=in_fpn, out_fpn=out_fpn)


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"
__maintainer__ = "Jianfeng Sun"

from functools import wraps
from umiche.plot.gadget.Draw import Draw as draw
from umiche.plot.gadget.Touch import Touch as touch


class Transmitter:

    def __init__(
            self,
            type,
            task=None,
    ):
        self.type = type
        self.task = task
        self.touch = touch()

    def __call__(self, deal):
        if self.type == "bar_individual":
            draw_canvas = self.bar_individual
        elif self.type == "hist_grouped":
            draw_canvas = self.hist_grouped
        elif self.type == "line":
            draw_canvas = self.line
        elif self.type == "line_scatter":
            draw_canvas = self.line_scatter
        if self.task == 'annotate_height':
            tinker = self.touch.bar_height_label
        else:
            tinker = None
        @wraps(deal)
        def canvas(self, *args, **kwargs):
            # print(kwargs)
            proc_f = draw_canvas(*args, **kwargs)
            # print(proc_f)
            if tinker:
                tinker(
                    canvas=proc_f['canvas'],
                    ax=proc_f['ax'],
                    format=proc_f['y_format'],
                    fontsize=10,
                )
            return deal(self, *args, **kwargs)
        return canvas

    @draw(type='bar_individual')
    def bar_individual(*args, **kwargs):
            return kwargs

    @draw(type='hist_grouped')
    def hist_grouped(*args, **kwargs):
        return kwargs

    @draw(type='line')
    def line(*args, **kwargs):
        return kwargs

    @draw(type='line_scatter')
    def line_scatter(*args, **kwargs):
        return kwargs